# Copyright (C) 2020 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast
import copy
from datetime import datetime
import functools
import hashlib
import inspect
import os
import re
import traceback

from oslo_config import cfg
from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import encodeutils
from oslo_utils import excutils
from oslo_utils import timeutils
from toscaparser import tosca_template

from tacker.common import driver_manager
from tacker.common import exceptions
from tacker.common import log
from tacker.common import safe_utils
from tacker.common import utils
from tacker.conductor.conductorrpc import vnf_lcm_rpc
from tacker import manager
from tacker import objects
from tacker.objects import fields
from tacker.objects.fields import ErrorPoint as EP
from tacker.vnflcm import abstract_driver
from tacker.vnflcm import utils as vnflcm_utils

LOG = logging.getLogger(__name__)
CONF = cfg.CONF
DEFAULT_VNFLCM_MGMT_DRIVER = "vnflcm_noop"


@utils.expects_func_args('vnf_info', 'vnf_instance', 'scale_vnf_request')
def revert_to_error_scale(function):
    """Decorator to revert task_state to error  on failure."""

    @functools.wraps(function)
    def decorated_function(self, context, *args, **kwargs):
        try:
            return function(self, context, *args, **kwargs)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                wrapped_func = safe_utils.get_wrapped_function(function)
                keyed_args = inspect.getcallargs(wrapped_func, self, context,
                                                 *args, **kwargs)
                try:
                    vnf_info = keyed_args['vnf_info']
                    vnf_instance = keyed_args['vnf_instance']
                    scale_vnf_request = keyed_args['scale_vnf_request']
                    vim_info = vnflcm_utils._get_vim(context,
                            vnf_instance.vim_connection_info)
                    vim_connection_info = \
                        objects.VimConnectionInfo.obj_from_primitive(
                            vim_info, context)
                    if vnf_info.get('resource_changes'):
                        resource_changes = vnf_info.get('resource_changes')
                    else:
                        resource_changes = self._scale_resource_update(context,
                                            vnf_info,
                                            vnf_instance,
                                            scale_vnf_request,
                                            vim_connection_info,
                                            error=True)
                except Exception as e:
                    LOG.warning(traceback.format_exc())
                    LOG.warning("Failed to scale resource update "
                                "instance %(id)s. Error: %(error)s",
                                {"id": vnf_instance.id, "error": e})

                try:
                    self._vnfm_plugin._update_vnf_scaling_status_err(context,
                                                                     vnf_info)
                except Exception as e:
                    LOG.warning("Failed to revert scale info for event "
                                "instance %(id)s. Error: %(error)s",
                                {"id": vnf_instance.id, "error": e})
                try:
                    vnf_instance.task_state = None
                    self._vnf_instance_update(context, vnf_instance)
                except Exception as e:
                    LOG.warning("Failed to revert instantiation info for vnf "
                                "instance %(id)s. Error: %(error)s",
                                {"id": vnf_instance.id, "error": e})
                problem = objects.ProblemDetails(status=500,
                                                 detail=str(ex))

                try:
                    timestamp = datetime.utcnow()
                    vnf_lcm_op_occ = vnf_info['vnf_lcm_op_occ']
                    vnf_lcm_op_occ.operation_state = 'FAILED_TEMP'
                    vnf_lcm_op_occ.state_entered_time = timestamp
                    vnf_lcm_op_occ.resource_changes = resource_changes
                    vnf_lcm_op_occ.error = problem
                    vnf_lcm_op_occ.error_point = \
                        vnf_info['current_error_point']
                    vnf_lcm_op_occ.save()
                except Exception as e:
                    LOG.warning("Failed to update vnf_lcm_op_occ for vnf "
                                "instance %(id)s. Error: %(error)s",
                                {"id": vnf_instance.id, "error": e})

                try:
                    notification = vnf_info['notification']
                    notification['notificationStatus'] = 'RESULT'
                    notification['operationState'] = 'FAILED_TEMP'
                    notification['error'] = problem.to_dict()
                    resource_dict = resource_changes.to_dict()
                    if resource_dict.get('affected_vnfcs'):
                        notification['affectedVnfcs'] =\
                            jsonutils.dump_as_bytes(
                            resource_dict.get('affected_vnfcs'))
                    if resource_dict.get('affected_virtual_links'):
                        notification['affectedVirtualLinks'] =\
                            jsonutils.dump_as_bytes(
                            resource_dict.get('affected_virtual_links'))
                    if resource_dict.get('affected_virtual_storages'):
                        notification['affectedVirtualStorages'] =\
                            jsonutils.dump_as_bytes(
                            resource_dict.get('affected_virtual_storages'))
                    self.rpc_api.send_notification(context, notification)
                except Exception as e:
                    LOG.warning("Failed to revert scale info for vnf "
                                "instance %(id)s. Error: %(error)s",
                                {"id": vnf_instance.id, "error": e})

    return decorated_function


@utils.expects_func_args('vnf_instance')
def revert_to_error_task_state(function):
    """Decorator to revert task_state to error  on failure."""

    @functools.wraps(function)
    def decorated_function(self, context, *args, **kwargs):
        try:
            return function(self, context, *args, **kwargs)
        except Exception:
            with excutils.save_and_reraise_exception():
                wrapped_func = safe_utils.get_wrapped_function(function)
                keyed_args = inspect.getcallargs(wrapped_func, self, context,
                                                 *args, **kwargs)
                vnf_instance = keyed_args['vnf_instance']
                previous_task_state = vnf_instance.task_state
                try:
                    self._vnf_instance_update(context, vnf_instance,
                        task_state=fields.VnfInstanceTaskState.ERROR)
                    LOG.info("Successfully reverted task state from "
                             "%(state)s to %(error)s on failure for vnf "
                             "instance %(id)s.",
                             {"state": previous_task_state,
                              "id": vnf_instance.id,
                              "error": fields.VnfInstanceTaskState.ERROR})
                except Exception as e:
                    LOG.warning("Failed to revert task state for vnf "
                                "instance %(id)s. Error: %(error)s",
                                {"id": vnf_instance.id, "error": e})

    return decorated_function


@utils.expects_func_args('vnf_info', 'vnf_instance', 'operation_params')
def revert_to_error_rollback(function):
    """Decorator to revert task_state to error  on failure."""

    @functools.wraps(function)
    def decorated_function(self, context, *args, **kwargs):
        try:
            return function(self, context, *args, **kwargs)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                wrapped_func = safe_utils.get_wrapped_function(function)
                keyed_args = inspect.getcallargs(wrapped_func, self, context,
                                                 *args, **kwargs)
                resource_changes = None
                try:
                    vnf_info = keyed_args['vnf_info']
                    vnf_instance = keyed_args['vnf_instance']
                    operation_params = keyed_args['operation_params']
                    vim_info = vnflcm_utils._get_vim(context,
                            vnf_instance.vim_connection_info)
                    vim_connection_info =\
                        objects.VimConnectionInfo.obj_from_primitive(
                            vim_info, context)
                    vnf_lcm_op_occs = vnf_info['vnf_lcm_op_occ']
                    if vnf_info.get('resource_changes'):
                        resource_changes = vnf_info.get('resource_changes')
                    else:
                        if vnf_lcm_op_occs.operation == 'SCALE':
                            scale_vnf_request =\
                                objects.ScaleVnfRequest.obj_from_primitive(
                                    operation_params, context=context)
                            scale_vnf_request_copy = \
                                copy.deepcopy(scale_vnf_request)
                            scale_vnf_request_copy.type = 'SCALE_IN'
                            resource_changes = self._scale_resource_update(
                                context,
                                vnf_info,
                                vnf_instance,
                                scale_vnf_request_copy,
                                vim_connection_info,
                                error=True)
                        else:
                            resource_changes = self._term_resource_update(
                                context,
                                vnf_info,
                                vnf_instance)
                except Exception as e:
                    LOG.warning(traceback.format_exc())
                    LOG.warning("Failed to scale resource update "
                                "instance %(id)s. Error: %(error)s",
                                {"id": vnf_instance.id, "error": e})

                try:
                    self._update_vnf_rollback_status_err(context, vnf_info)
                except Exception as e:
                    LOG.warning("Failed to revert scale info for event "
                                "instance %(id)s. Error: %(error)s",
                                {"id": vnf_instance.id, "error": e})
                try:
                    self._vnf_instance_update(context, vnf_instance)
                except Exception as e:
                    LOG.warning("Failed to revert instantiation info for vnf "
                                "instance %(id)s. Error: %(error)s",
                                {"id": vnf_instance.id, "error": e})
                problem = objects.ProblemDetails(status=500,
                                                 detail=str(ex))

                try:
                    timestamp = datetime.utcnow()
                    vnf_lcm_op_occ = vnf_info['vnf_lcm_op_occ']
                    vnf_lcm_op_occ.operation_state = 'FAILED_TEMP'
                    vnf_lcm_op_occ.state_entered_time = timestamp
                    if resource_changes:
                        vnf_lcm_op_occ.resource_changes = resource_changes
                    vnf_lcm_op_occ.error = problem
                    vnf_lcm_op_occ.save()
                except Exception as e:
                    LOG.warning("Failed to update vnf_lcm_op_occ for vnf "
                                "instance %(id)s. Error: %(error)s",
                                {"id": vnf_instance.id, "error": e})

                try:
                    notification = vnf_info['notification']
                    notification['notificationStatus'] = 'RESULT'
                    notification['operationState'] = 'FAILED_TEMP'
                    notification['error'] = problem.to_dict()
                    if resource_changes:
                        resource_dict = resource_changes.to_dict()
                        if resource_dict.get('affected_vnfcs'):
                            notification['affectedVnfcs'] = \
                                jsonutils.dump_as_bytes(
                                    resource_dict.get('affected_vnfcs'))
                        if resource_dict.get('affected_virtual_links'):
                            notification['affectedVirtualLinks'] = \
                                jsonutils.dump_as_bytes(
                                    resource_dict.get(
                                        'affected_virtual_links'))
                        if resource_dict.get('affected_virtual_storages'):
                            notification['affectedVirtualStorages'] = \
                                jsonutils.dump_as_bytes(
                                    resource_dict.get(
                                        'affected_virtual_storages'))
                    self.rpc_api.send_notification(context, notification)
                except Exception as e:
                    LOG.warning("Failed to revert scale info for vnf "
                                "instance %(id)s. Error: %(error)s",
                                {"id": vnf_instance.id, "error": e})
    return decorated_function


def config_opts():
    return [('tacker', VnfLcmDriver.OPTS)]


class VnfLcmDriver(abstract_driver.VnfInstanceAbstractDriver):
    OPTS = [
        cfg.ListOpt(
            'vnflcm_infra_driver', default=['openstack', 'kubernetes'],
            help=_('Hosting vnf drivers tacker plugin will use')
        ),
        cfg.ListOpt(
            'vnflcm_mgmt_driver', default=[DEFAULT_VNFLCM_MGMT_DRIVER],
            help=_('MGMT driver to communicate with '
                   'Hosting VNF/logical service '
                   'instance tacker plugin will use')
        )
    ]
    cfg.CONF.register_opts(OPTS, 'tacker')

    def __init__(self):
        super(VnfLcmDriver, self).__init__()
        self.rpc_api = vnf_lcm_rpc.VNFLcmRPCAPI()
        self._vnfm_plugin = manager.TackerManager.get_service_plugins()['VNFM']
        self._vnf_manager = driver_manager.DriverManager(
            'tacker.tacker.vnfm.drivers',
            cfg.CONF.tacker.vnflcm_infra_driver)
        self._mgmt_manager = driver_manager.DriverManager(
            'tacker.tacker.mgmt.drivers', cfg.CONF.tacker.vnflcm_mgmt_driver)
        self._mgmt_driver_hash = self._init_mgmt_driver_hash()

    def _init_mgmt_driver_hash(self):

        driver_hash = {}
        for mgmt_driver in cfg.CONF.tacker.vnflcm_mgmt_driver:
            path = inspect.getfile(self._mgmt_manager[mgmt_driver].__class__)
            driver_hash[mgmt_driver] = self._get_file_hash(path)
        return driver_hash

    def _vnf_instance_update(self, context, vnf_instance, **kwargs):
        """Update vnf instance in the database using kwargs as value."""

        for k, v in kwargs.items():
            setattr(vnf_instance, k, v)
        vnf_instance.save()

    def _instantiate_vnf(self, context, vnf_instance, vnf_dict,
            vim_connection_info, instantiate_vnf_req):
        vnfd_dict = vnflcm_utils._get_vnfd_dict(context, vnf_instance.vnfd_id,
                instantiate_vnf_req.flavour_id)
        vnf_package_path = vnflcm_utils._get_vnf_package_path(
            context, vnf_instance.vnfd_id)

        param_for_subs_map = vnflcm_utils._get_param_data(vnfd_dict,
                instantiate_vnf_req)

        package_uuid = vnflcm_utils._get_vnf_package_id(context,
                vnf_instance.vnfd_id)
        vnf_software_images = vnflcm_utils._create_grant_request(vnfd_dict,
                package_uuid)
        vnf_resources = self._vnf_manager.invoke(
            vim_connection_info.vim_type, 'pre_instantiation_vnf',
            context=context, vnf_instance=vnf_instance,
            vim_connection_info=vim_connection_info,
            vnf_software_images=vnf_software_images,
            instantiate_vnf_req=instantiate_vnf_req,
            vnf_package_path=vnf_package_path)

        # save the vnf resources in the db
        for _, resources in vnf_resources.items():
            for vnf_resource in resources:
                vnf_resource.create()

        vnfd_dict_to_create_final_dict = copy.deepcopy(vnfd_dict)
        final_vnf_dict = vnflcm_utils._make_final_vnf_dict(
            vnfd_dict_to_create_final_dict, vnf_instance.id,
            vnf_instance.vnf_instance_name, param_for_subs_map, vnf_dict)

        final_vnf_dict['before_error_point'] = \
            vnf_dict['before_error_point']

        try:
            instance_id = self._vnf_manager.invoke(
                vim_connection_info.vim_type, 'instantiate_vnf',
                context=context, plugin=self._vnfm_plugin,
                vnf_instance=vnf_instance,
                vnfd_dict=final_vnf_dict, grant_response=vnf_resources,
                vim_connection_info=vim_connection_info,
                vnf_package_path=vnf_package_path,
                instantiate_vnf_req=instantiate_vnf_req)
        except Exception as exp:
            with excutils.save_and_reraise_exception():
                exp.reraise = False
                LOG.error("Unable to instantiate vnf instance "
                    "%(id)s due to error : %(error)s",
                    {"id": vnf_instance.id, "error":
                    encodeutils.exception_to_unicode(exp)})
                raise exceptions.VnfInstantiationFailed(
                    id=vnf_instance.id,
                    error=encodeutils.exception_to_unicode(exp))

        if vnf_instance.instantiated_vnf_info and\
                vnf_instance.instantiated_vnf_info.instance_id != instance_id:
            # TODO(h-asahina): rename instance_id to stack_id
            vnf_instance.instantiated_vnf_info.instance_id = instance_id

        if vnf_dict['attributes'].get('scaling_group_names'):
            vnf_instance.instantiated_vnf_info.scale_status = \
                vnf_dict['scale_status']
        elif vnf_instance.instantiated_vnf_info:
            default_scale_status = vnflcm_utils.\
                get_default_scale_status(
                    context=context,
                    vnf_instance=vnf_instance,
                    vnfd_dict=vnfd_dict)
            if default_scale_status is not None:
                vnf_instance.instantiated_vnf_info.scale_status = \
                    default_scale_status

        if vnf_dict['before_error_point'] <= EP.PRE_VIM_CONTROL:
            try:
                self._vnf_manager.invoke(
                    vim_connection_info.vim_type, 'create_wait',
                    plugin=self._vnfm_plugin, context=context,
                    vnf_dict=final_vnf_dict,
                    vnf_id=final_vnf_dict['instance_id'],
                    auth_attr=vim_connection_info.access_info)

            except Exception as exp:
                with excutils.save_and_reraise_exception():
                    exp.reraise = False
                    LOG.error("Vnf creation wait failed for vnf instance "
                        "%(id)s due to error : %(error)s",
                        {"id": vnf_instance.id, "error":
                        encodeutils.exception_to_unicode(exp)})
                    raise exceptions.VnfInstantiationWaitFailed(
                        id=vnf_instance.id,
                        error=encodeutils.exception_to_unicode(exp))
        elif vnf_dict['before_error_point'] == EP.POST_VIM_CONTROL:
            try:
                self._vnf_manager.invoke(
                    vim_connection_info.vim_type, 'update_stack_wait',
                    plugin=self._vnfm_plugin, context=context,
                    vnf_dict=final_vnf_dict,
                    stack_id=instance_id,
                    auth_attr=vim_connection_info.access_info)

            except Exception as exp:
                with excutils.save_and_reraise_exception():
                    exp.reraise = False
                    LOG.error("Vnf update wait failed for vnf instance "
                        "%(id)s due to error : %(error)s",
                        {"id": vnf_instance.id, "error":
                        encodeutils.exception_to_unicode(exp)})
                    raise exceptions.VnfInstantiationWaitFailed(
                        id=vnf_instance.id,
                        error=encodeutils.exception_to_unicode(exp))

        if vnf_instance.instantiated_vnf_info.instance_id:
            self._vnf_manager.invoke(vim_connection_info.vim_type,
                                    'post_vnf_instantiation', context=context,
                                    vnf_instance=vnf_instance,
                                    vim_connection_info=vim_connection_info,
                                    instantiate_vnf_req=instantiate_vnf_req)

    def _get_file_hash(self, path):
        hash_obj = hashlib.sha256()
        with open(path) as f:
            hash_obj.update(f.read().encode('utf-8'))
        return hash_obj.hexdigest()

    def _check_mgmt_driver(self, artifact_mgmt_driver, artifacts_value,
                           vnf_package_path):
        # check implementation and artifacts exist in cfg.CONF.tacker
        if artifact_mgmt_driver not in self._mgmt_driver_hash:
            LOG.error('The {} specified in the VNFD '
                      'is inconsistent with the MgmtDriver in '
                      'the configuration file.'.format(artifact_mgmt_driver))
            raise exceptions.MgmtDriverInconsistent(
                MgmtDriver=artifact_mgmt_driver)

        # check file content
        pkg_mgmt_driver_path = os.path.join(vnf_package_path,
            artifacts_value[artifact_mgmt_driver]['file'])
        pkg_mgmt_driver_hash = self._get_file_hash(pkg_mgmt_driver_path)
        if pkg_mgmt_driver_hash == \
                self._mgmt_driver_hash[artifact_mgmt_driver]:
            return artifact_mgmt_driver
        else:
            LOG.error('The hash verification of VNF Package MgmtDriver '
                      'and Tacker MgmtDriver does not match.')
            raise exceptions.MgmtDriverHashMatchFailure()

    def _load_vnf_interface(self, context, method_name,
                            vnf_instance, vnfd_dict):
        VNF_value = vnfd_dict['topology_template']['node_templates']['VNF']
        tacker_mgmt_driver = DEFAULT_VNFLCM_MGMT_DRIVER
        interfaces_vnflcm_value = \
            VNF_value.get('interfaces', {}).get('Vnflcm', {})
        if not interfaces_vnflcm_value:
            return tacker_mgmt_driver
        artifacts_value = VNF_value.get('artifacts')
        if not artifacts_value:
            return tacker_mgmt_driver
        vnf_package_path = vnflcm_utils._get_vnf_package_path(
            context, vnf_instance.vnfd_id)

        if interfaces_vnflcm_value.get(method_name):
            artifact_mgmt_driver = interfaces_vnflcm_value.get(
                method_name).get('implementation')
            if artifact_mgmt_driver:
                tacker_mgmt_driver = self._check_mgmt_driver(
                    artifact_mgmt_driver, artifacts_value, vnf_package_path)

        return tacker_mgmt_driver

    @log.log
    def modify_vnf(self, context, vnf_lcm_opoccs,
                   body_data, vnfd_pkg_data, vnfd_id):
        # Get vnf_instance from DB by vnf_instance_id
        vnf_instance_id = vnf_lcm_opoccs.get('vnf_instance_id')
        vnf_instance = objects.VnfInstance.get_by_id(context, vnf_instance_id)
        vnfd_dict = vnflcm_utils.get_vnfd_dict(
            context, vnf_instance.vnfd_id,
            vnf_instance.instantiated_vnf_info.flavour_id)

        # modify_information_start(context, vnf_instance)
        self._mgmt_manager.invoke(
            self._load_vnf_interface(
                context, 'modify_information_start', vnf_instance, vnfd_dict),
            'modify_information_start', context=context,
            modify_vnf_request=None, vnf_instance=vnf_instance)

        # Get the old vnf package path according to vnfd_id
        old_vnf_package_path = vnflcm_utils.get_vnf_package_path(
            context, vnf_instance.vnfd_id)

        # Update vnf_instance
        try:
            _vnf_instance = objects.VnfInstance(context=context)
            updated_time = _vnf_instance.update(
                context, vnf_lcm_opoccs, body_data, vnfd_pkg_data, vnfd_id)
        except Exception as msg:
            raise Exception(str(msg))
        vnf_instance = objects.VnfInstance.get_by_id(context, vnf_instance_id)

        vim_info = vnflcm_utils.get_vim(
            context, vnf_instance.vim_connection_info)

        vim_connection_info = objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context)

        kwargs = {}
        if vim_connection_info.vim_type == 'kubernetes':
            # If the file path of ConfigMap/Secret is changed
            cm_secret_paths = []
            # Get the metadata from vnf_lcm_opoccs
            operation_params = vnf_lcm_opoccs.get('operationParams')
            if operation_params:
                try:
                    cm_secret_paths = (ast.literal_eval(operation_params)
                                       .get('metadata', {})
                                       .get('configmap_secret_paths', []))
                except Exception as e:
                    LOG.error('Invalid format operationParams')
                    raise exceptions.InvalidInput(str(e))
            kwargs = {"old_vnf_package_path": old_vnf_package_path,
                      "configmap_secret_paths": cm_secret_paths}

        self._mgmt_manager.invoke(
            self._load_vnf_interface(
                context, 'modify_information_end', vnf_instance, vnfd_dict),
            'modify_information_end', context=context,
            modify_vnf_request=None,
            vnf_instance=vnf_instance, **kwargs)
        return updated_time

    @log.log
    def instantiate_vnf(self, context, vnf_instance, vnf_dict,
            instantiate_vnf_req):

        vnf_dict['current_error_point'] = EP.VNF_CONFIG_START
        vim_connection_info_list = vnflcm_utils.\
            _get_vim_connection_info_from_vnf_req(vnf_instance,
                    instantiate_vnf_req)

        self._vnf_instance_update(context, vnf_instance,
                vim_connection_info=vim_connection_info_list)

        vim_info = vnflcm_utils._get_vim(context,
                instantiate_vnf_req.vim_connection_info)

        if vim_info['tenant_id'] != vnf_instance.tenant_id:
            LOG.error('The target VNF %(id)s cannot be instantiate '
                      'from a VIM of a different tenant.',
                      {"id": vnf_instance.id})
            raise exceptions.TenantMatchFailure(resource='VNF',
                id=vnf_instance.id,
                action='instantiate')

        vim_connection_info = objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context)

        vnfd_dict = vnflcm_utils._get_vnfd_dict(
            context, vnf_instance.vnfd_id, instantiate_vnf_req.flavour_id)

        if vnf_dict['before_error_point'] <= EP.VNF_CONFIG_START:
            # TODO(LiangLu): grant_request here is planned to pass
            # as a parameter, however due to grant_request is not
            # passed from conductor to vnflcm_driver, thus we put Null
            # value to grant_reqeust temporary.
            # This part will be updated in next release.
            self._mgmt_manager.invoke(
                self._load_vnf_interface(
                    context, 'instantiate_start', vnf_instance, vnfd_dict),
                'instantiate_start', context=context,
                vnf_instance=vnf_instance,
                instantiate_vnf_request=instantiate_vnf_req,
                grant=vnf_dict.get('grant'), grant_request=None)

        vnf_dict['current_error_point'] = EP.PRE_VIM_CONTROL
        if vnf_dict['before_error_point'] <= EP.POST_VIM_CONTROL:
            try:
                self._instantiate_vnf(context, vnf_instance, vnf_dict,
                                      vim_connection_info, instantiate_vnf_req)
            except Exception as exc:
                if (hasattr(vnf_instance.instantiated_vnf_info, 'instance_id')
                        and vnf_instance.instantiated_vnf_info.instance_id):
                    vnf_dict['current_error_point'] = EP.POST_VIM_CONTROL
                raise exc

        vnf_dict['current_error_point'] = EP.INTERNAL_PROCESSING
        vnf_dict['current_error_point'] = EP.VNF_CONFIG_END
        if vnf_dict['before_error_point'] <= EP.VNF_CONFIG_END:
            # TODO(LiangLu): grant_request here is planned to pass
            # as a parameter, however due to grant_request is not
            # passed from conductor to vnflcm_driver, thus we put Null
            # value to grant_reqeust temporary.
            # This part will be updated in next release.
            kwargs = {'vnf': copy.deepcopy(vnf_dict)}
            self._mgmt_manager.invoke(
                self._load_vnf_interface(
                    context, 'instantiate_end', vnf_instance, vnfd_dict),
                'instantiate_end', context=context,
                vnf_instance=vnf_instance,
                instantiate_vnf_request=instantiate_vnf_req,
                grant=vnf_dict.get('grant'), grant_request=None, **kwargs)

    @log.log
    @revert_to_error_task_state
    def terminate_vnf(self, context, vnf_instance, terminate_vnf_req,
            vnf_dict):
        vnf_dict['current_error_point'] = EP.VNF_CONFIG_START

        vim_info = vnflcm_utils._get_vim(context,
            vnf_instance.vim_connection_info)

        vim_connection_info = objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context)

        vnfd_dict = vnflcm_utils._get_vnfd_dict(
            context, vnf_instance.vnfd_id,
            vnf_instance.instantiated_vnf_info.flavour_id)

        if vnf_dict['before_error_point'] <= EP.VNF_CONFIG_START:
            # TODO(LiangLu): grant_request and grant here is planned to
            # pass as a parameter, however due to they are not
            # passed from conductor to vnflcm_driver, thus we put Null
            # value to grant and grant_reqeust temporary.
            # This part will be updated in next release.
            kwargs = {'vnf': copy.deepcopy(vnf_dict)}
            self._mgmt_manager.invoke(
                self._load_vnf_interface(
                    context, 'terminate_start', vnf_instance, vnfd_dict),
                'terminate_start', context=context,
                vnf_instance=vnf_instance,
                terminate_vnf_request=terminate_vnf_req,
                grant=None, grant_request=None, **kwargs)

        vnf_dict['current_error_point'] = EP.PRE_VIM_CONTROL

        LOG.info("Terminating vnf %s", vnf_instance.id)
        try:
            if vnf_dict['before_error_point'] <= EP.POST_VIM_CONTROL:
                self._delete_vnf_instance_resources(context, vnf_instance,
                        vim_connection_info,
                        terminate_vnf_req=terminate_vnf_req)

            vnf_dict['current_error_point'] = EP.INTERNAL_PROCESSING
            vnf_instance.instantiated_vnf_info.reinitialize()
            self._vnf_instance_update(context, vnf_instance,
                vim_connection_info=[], task_state=None)

            LOG.info("Vnf terminated %s successfully", vnf_instance.id)
        except Exception as exp:
            with excutils.save_and_reraise_exception():
                if vnf_dict['current_error_point'] == EP.PRE_VIM_CONTROL:
                    if hasattr(vnf_instance.instantiated_vnf_info,
                            'instance_id'):
                        if vnf_instance.instantiated_vnf_info.instance_id:
                            vnf_dict['current_error_point'] = \
                                EP.POST_VIM_CONTROL

                LOG.error("Unable to terminate vnf '%s' instance. "
                          "Error: %s", vnf_instance.id,
                          encodeutils.exception_to_unicode(exp))

        vnf_dict['current_error_point'] = EP.VNF_CONFIG_END
        if vnf_dict['before_error_point'] <= EP.VNF_CONFIG_END:
            # TODO(LiangLu): grant_request and grant here is planned to
            # pass as a parameter, however due to they are not
            # passed from conductor to vnflcm_driver, thus we put Null
            # value to grant and grant_reqeust temporary.
            # This part will be updated in next release.
            self._mgmt_manager.invoke(
                self._load_vnf_interface(
                    context, 'terminate_end', vnf_instance, vnfd_dict),
                'terminate_end', context=context,
                vnf_instance=vnf_instance,
                terminate_vnf_request=terminate_vnf_req,
                grant=None, grant_request=None)

    def _delete_vnf_instance_resources(self, context, vnf_instance,
            vim_connection_info, terminate_vnf_req=None,
            update_instantiated_state=True):

        if (vnf_instance.instantiated_vnf_info and
            vnf_instance.instantiated_vnf_info.instance_id) or \
                vim_connection_info.vim_type == 'kubernetes':

            instance_id = vnf_instance.instantiated_vnf_info.instance_id \
                if vnf_instance.instantiated_vnf_info else None
            access_info = vim_connection_info.access_info

            LOG.info("Deleting stack %(instance)s for vnf %(id)s ",
                    {"instance": instance_id, "id": vnf_instance.id})

            self._vnf_manager.invoke(vim_connection_info.vim_type,
                'delete', plugin=self, context=context,
                vnf_id=instance_id, auth_attr=access_info,
                vnf_instance=vnf_instance, terminate_vnf_req=terminate_vnf_req)

            if update_instantiated_state:
                vnf_instance.instantiation_state = \
                    fields.VnfInstanceState.NOT_INSTANTIATED
                vnf_instance.save()

            self._vnf_manager.invoke(vim_connection_info.vim_type,
                'delete_wait', plugin=self, context=context,
                vnf_id=instance_id, auth_attr=access_info,
                vnf_instance=vnf_instance)

        vnf_resources = objects.VnfResourceList.get_by_vnf_instance_id(
            context, vnf_instance.id)

        for vnf_resource in vnf_resources:
            self._vnf_manager.invoke(vim_connection_info.vim_type,
                    'delete_vnf_instance_resource',
                    context=context, vnf_instance=vnf_instance,
                    vim_connection_info=vim_connection_info,
                    vnf_resource=vnf_resource)

            vnf_resource.destroy(context)

    def _heal_vnf(self, context, vnf_instance, vim_connection_info,
            heal_vnf_request, vnf):
        inst_vnf_info = vnf_instance.instantiated_vnf_info
        try:
            self._vnf_manager.invoke(
                vim_connection_info.vim_type, 'heal_vnf',
                context=context, vnf_instance=vnf_instance,
                vim_connection_info=vim_connection_info,
                heal_vnf_request=heal_vnf_request)
        except Exception as exp:
            with excutils.save_and_reraise_exception() as exc_ctxt:
                exc_ctxt.reraise = False
                LOG.error("Failed to heal vnf %(id)s in infra driver. "
                          "Error: %(error)s", {"id": vnf_instance.id, "error":
                          encodeutils.exception_to_unicode(exp)})
                raise exceptions.VnfHealFailed(id=vnf_instance.id,
                    error=encodeutils.exception_to_unicode(exp))

        vnf['current_error_point'] = EP.POST_VIM_CONTROL

        try:
            self._vnf_manager.invoke(
                vim_connection_info.vim_type, 'heal_vnf_wait',
                context=context, vnf_instance=vnf_instance,
                vim_connection_info=vim_connection_info,
                heal_vnf_request=heal_vnf_request)
        except Exception as exp:
            LOG.error("Failed to update vnf %(id)s resources for instance "
                      "%(instance)s. Error: %(error)s",
                      {'id': vnf_instance.id, 'instance':
                      inst_vnf_info.instance_id, 'error':
                      encodeutils.exception_to_unicode(exp)})

        try:
            self._vnf_manager.invoke(
                vim_connection_info.vim_type, 'post_heal_vnf',
                context=context, vnf_instance=vnf_instance,
                vim_connection_info=vim_connection_info,
                heal_vnf_request=heal_vnf_request)
            self._vnf_instance_update(context, vnf_instance, task_state=None)
        except Exception as exp:
            with excutils.save_and_reraise_exception() as exc_ctxt:
                exc_ctxt.reraise = False
                LOG.error("Failed to store updated resources information for "
                          "instance %(instance)s for vnf %(id)s. "
                          "Error: %(error)s",
                          {'id': vnf_instance.id, 'instance':
                          inst_vnf_info.instance_id, 'error':
                          encodeutils.exception_to_unicode(exp)})
                raise exceptions.VnfHealFailed(id=vnf_instance.id,
                    error=encodeutils.exception_to_unicode(exp))

    def _respawn_vnf(self, context, vnf_instance, vnf_dict,
                    vim_connection_info, heal_vnf_request):
        if vnf_dict['before_error_point'] != EP.POST_VIM_CONTROL:
            try:
                self._delete_vnf_instance_resources(context, vnf_instance,
                    vim_connection_info, update_instantiated_state=False)
            except Exception as exc:
                with excutils.save_and_reraise_exception() as exc_ctxt:
                    exc_ctxt.reraise = False
                    err_msg = ("Failed to delete vnf resources for "
                              "vnf instance %(id)s before respawning. "
                              "The vnf is in inconsistent state. "
                              "Error: %(error)s")
                    LOG.error(err_msg % {"id": vnf_instance.id,
                              "error": str(exc)})
                    raise exceptions.VnfHealFailed(id=vnf_instance.id,
                        error=encodeutils.exception_to_unicode(exc))

        # InstantiateVnfRequest is not stored in the db as it's mapped
        # to InstantiatedVnfInfo version object. Convert InstantiatedVnfInfo
        # version object to InstantiateVnfRequest so that vnf can be
        # instantiated.

        instantiate_vnf_request = objects.InstantiateVnfRequest.\
            from_vnf_instance(vnf_instance)
        vnf_instance.instantiated_vnf_info.reinitialize()
        vnf_instance.task_state = fields.VnfInstanceTaskState.INSTANTIATING
        vnfd_dict = vnflcm_utils._get_vnfd_dict(
            context, vnf_instance.vnfd_id, instantiate_vnf_request.flavour_id)
        if vnf_dict.get('vnf_instance_after'):
            vnf_instance.instantiated_vnf_info = \
                vnf_dict.get('vnf_instance_after').instantiated_vnf_info
        else:
            vnflcm_utils._build_instantiated_vnf_info(
                vnfd_dict, instantiate_vnf_request, vnf_instance,
                vim_connection_info.vim_id)

        try:
            self._instantiate_vnf(context, vnf_instance, vnf_dict,
                                  vim_connection_info, instantiate_vnf_request)
            self._vnf_manager.invoke(
                vim_connection_info.vim_type, 'post_vnf_instantiation',
                context=context, vnf_instance=vnf_instance,
                vim_connection_info=vim_connection_info,
                instantiate_vnf_req=instantiate_vnf_request)

        except Exception as exc:
            with excutils.save_and_reraise_exception() as exc_ctxt:
                exc_ctxt.reraise = False
                err_msg = ("Failed to instantiate vnf instance "
                          "%(id)s after termination. The vnf is in "
                          "inconsistent state. Error: %(error)s")
                LOG.error(err_msg % {"id": vnf_instance.id,
                          "error": str(exc)})
                raise exceptions.VnfHealFailed(id=vnf_instance.id,
                    error=encodeutils.exception_to_unicode(exc))

        self._vnf_instance_update(context, vnf_instance,
                    instantiation_state=fields.VnfInstanceState.INSTANTIATED,
                    task_state=None)

    @log.log
    @revert_to_error_task_state
    def heal_vnf(self, context, vnf_instance, vnf_dict, heal_vnf_request):
        vnf_dict['current_error_point'] = EP.VNF_CONFIG_START

        LOG.info("Request received for healing vnf '%s'", vnf_instance.id)
        vim_info = vnflcm_utils._get_vim(context,
            vnf_instance.vim_connection_info)

        vim_connection_info = objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context)

        vnfd_dict = vnflcm_utils._get_vnfd_dict(
            context, vnf_instance.vnfd_id,
            vnf_instance.instantiated_vnf_info.flavour_id)

        if vnf_dict['before_error_point'] <= EP.VNF_CONFIG_START:
            # TODO(LiangLu): grant_request here is planned to pass
            # as a parameter, however due to grant_request are not
            # passed from conductor to vnflcm_driver, thus we put Null
            # value to grant and grant_reqeust temporary.
            # This part will be updated in next release.
            self._mgmt_manager.invoke(
                self._load_vnf_interface(
                    context, 'heal_start', vnf_instance, vnfd_dict),
                'heal_start', context=context,
                vnf_instance=vnf_instance,
                heal_vnf_request=heal_vnf_request,
                grant=vnf_dict.get('grant'), grant_request=None)

        vnf_dict['current_error_point'] = EP.PRE_VIM_CONTROL

        try:
            heal_flag = False
            if vnf_dict['before_error_point'] <= EP.POST_VIM_CONTROL:
                if not heal_vnf_request.vnfc_instance_id:
                    self._respawn_vnf(context, vnf_instance, vnf_dict,
                                    vim_connection_info, heal_vnf_request)
                else:
                    heal_flag = True
                    self._heal_vnf(context, vnf_instance, vim_connection_info,
                            heal_vnf_request, vnf_dict)

                LOG.info("Request received for healing vnf '%s' is completed "
                        "successfully", vnf_instance.id)
        except Exception as exp:
            with excutils.save_and_reraise_exception():
                if vnf_dict['current_error_point'] == EP.PRE_VIM_CONTROL:
                    if not heal_flag:
                        if hasattr(vnf_instance.instantiated_vnf_info,
                                'instance_id'):
                            if vnf_instance.instantiated_vnf_info.instance_id:
                                vnf_dict['current_error_point'] = \
                                    EP.POST_VIM_CONTROL

                LOG.error("Unable to heal vnf '%s' instance. "
                          "Error: %s", heal_vnf_request.vnfc_instance_id,
                          encodeutils.exception_to_unicode(exp))

                raise exceptions.VnfHealFailed(id=vnf_instance.id,
                    error=encodeutils.exception_to_unicode(exp))

        vnf_dict['current_error_point'] = EP.VNF_CONFIG_END

        if vnf_dict['before_error_point'] <= EP.VNF_CONFIG_END:
            # TODO(LiangLu): grant_request here is planned to pass
            # as a parameter, however due to grant_request are not
            # passed from conductor to vnflcm_driver, thus we put Null
            # value to grant and grant_reqeust temporary.
            # This part will be updated in next release.
            kwargs = {'vnf': copy.deepcopy(vnf_dict)}
            self._mgmt_manager.invoke(
                self._load_vnf_interface(
                    context, 'heal_end', vnf_instance, vnfd_dict),
                'heal_end', context=context,
                vnf_instance=vnf_instance,
                heal_vnf_request=heal_vnf_request,
                grant=vnf_dict.get('grant'), grant_request=None, **kwargs)

    def _scale_vnf_pre(self, context, vnf_info, vnf_instance,
                      scale_vnf_request, vim_connection_info):
        if vnf_info['before_error_point'] <= EP.NOTIFY_PROCESSING:
            self._vnfm_plugin._update_vnf_scaling(
                context, vnf_info, 'ACTIVE', 'PENDING_'
                + scale_vnf_request.type)
        vnf_info['current_error_point'] = EP.VNF_CONFIG_START

        scale_name_list = []
        grp_id = None
        vnf_info['policy_name'] = scale_vnf_request.aspect_id
        if scale_vnf_request.type == 'SCALE_IN':
            vnfd_dict = vnflcm_utils._get_vnfd_dict(
                context, vnf_instance.vnfd_id,
                vnf_instance.instantiated_vnf_info.flavour_id)

            vnf_info['action'] = 'in'
            scale_id_list, scale_name_list, grp_id, res_num = \
                self._vnf_manager.invoke(
                    vim_connection_info.vim_type,
                    'get_scale_in_ids',
                    plugin=self,
                    context=context,
                    vnf_dict=vnf_info,
                    is_reverse=scale_vnf_request.additional_params.get('\
                        is_reverse'),
                    auth_attr=vim_connection_info.access_info,
                    region_name=vim_connection_info.access_info.get('\
                        region_name'),
                    number_of_steps=scale_vnf_request.number_of_steps
                )
            vnf_info['res_num'] = res_num

            if vnf_info['before_error_point'] <= EP.VNF_CONFIG_START:
                # TODO(LiangLu): grant_request here is planned to pass
                # as a parameter, however due to grant_request are not
                # passed from conductor to vnflcm_driver, thus we put Null
                # value to grant and grant_reqeust temporary.
                # This part will be updated in next release.
                if len(scale_id_list) != 0 or \
                        vim_connection_info.vim_type == 'kubernetes':
                    kwargs = {'scale_name_list': scale_name_list,
                              'scale_stack_id': scale_id_list,
                              'vnf': copy.deepcopy(vnf_info)}
                    self._mgmt_manager.invoke(
                        self._load_vnf_interface(
                            context, 'scale_start', vnf_instance, vnfd_dict),
                        'scale_start', context=context,
                        vnf_instance=vnf_instance,
                        scale_vnf_request=scale_vnf_request,
                        grant=vnf_info.get('grant'), grant_request=None,
                        **kwargs)
        elif scale_vnf_request.type == 'SCALE_OUT':
            vnf_info['action'] = 'out'
            scale_id_list = self._vnf_manager.invoke(
                vim_connection_info.vim_type,
                'get_scale_ids',
                plugin=self,
                context=context,
                vnf_dict=vnf_info,
                auth_attr=vim_connection_info.access_info,
                region_name=vim_connection_info.access_info.get('region_name')
            )
        else:
            msg = 'Unknown vim type: %s' % vim_connection_info.vim_type
            raise exceptions.VnfScaleFailed(id=vnf_info['instance_id'],
                                            error=msg)

        vnf_info['current_error_point'] = EP.PRE_VIM_CONTROL
        return scale_id_list, scale_name_list, grp_id

    def _get_node_template_for_vnf(self, vnfd_dict):
        node_tmp = vnfd_dict['topology_template']['node_templates']
        for node_template in node_tmp.values():
            LOG.debug("node_template %s", node_template)
            if not re.match('^tosca', node_template['type']):
                LOG.debug("VNF node_template %s", node_template)
                return node_template
        return {}

    def _scale_vnf_post(self, context, vnf_info, vnf_instance,
                       scale_vnf_request, vim_connection_info,
                       scale_id_list,
                       resource_changes):
        vnf_info['current_error_point'] = EP.VNF_CONFIG_END
        if scale_vnf_request.type == 'SCALE_OUT':
            vnfd_dict = vnflcm_utils._get_vnfd_dict(
                context, vnf_instance.vnfd_id,
                vnf_instance.instantiated_vnf_info.flavour_id)

            scale_id_after = self._vnf_manager.invoke(
                vim_connection_info.vim_type,
                'get_scale_ids',
                plugin=self,
                context=context,
                vnf_dict=vnf_info,
                auth_attr=vim_connection_info.access_info,
                region_name=vim_connection_info.access_info.get('region_name')
            )
            id_list = list(set(scale_id_after) - set(scale_id_list))
            vnf_info['res_num'] = len(scale_id_after)

            if vnf_info['before_error_point'] <= EP.VNF_CONFIG_END:
                # TODO(LiangLu): grant_request here is planned to pass
                # as a parameter, however due to grant_request are not
                # passed from conductor to vnflcm_driver, thus we put Null
                # value to grant and grant_reqeust temporary.
                # This part will be updated in next release.
                if len(id_list) != 0 or \
                        vim_connection_info.vim_type == 'kubernetes':
                    kwargs = {'scale_stack_id': id_list,
                              'vnf': copy.deepcopy(vnf_info)}
                    self._mgmt_manager.invoke(
                        self._load_vnf_interface(
                            context, 'scale_end', vnf_instance, vnfd_dict),
                        'scale_end', context=context,
                        vnf_instance=vnf_instance,
                        scale_vnf_request=scale_vnf_request,
                        grant=vnf_info.get('grant'), grant_request=None,
                        **kwargs)

        vnf_instance.instantiated_vnf_info.scale_level =\
            vnf_info['after_scale_level']
        if vim_connection_info.vim_type != 'kubernetes':
            # NOTE(ueha): The logic of Scale for OpenStack VIM is widely hard
            # coded with `vnf_info`. This dependency is to be refactored in
            # future.
            scaleGroupDict = \
                jsonutils.loads(vnf_info['attributes']['scale_group'])
            (scaleGroupDict
            ['scaleGroupDict'][scale_vnf_request.aspect_id]['default']) =\
                vnf_info['res_num']
            vnf_info['attributes']['scale_group'] =\
                jsonutils.dump_as_bytes(scaleGroupDict)

        if vnf_info['before_error_point'] < EP.NOTIFY_COMPLETED:
            self._vnfm_plugin._update_vnf_scaling(context, vnf_info,
                    'PENDING_' + scale_vnf_request.type, 'ACTIVE')
            vnf_lcm_op_occ = vnf_info['vnf_lcm_op_occ']
            vnf_lcm_op_occ.operation_state = 'COMPLETED'
            vnf_lcm_op_occ.resource_changes = resource_changes
            vnf_lcm_op_occ.state_entered_time = timeutils.utcnow()
            vnf_lcm_op_occ.save()
            vnf_instance.task_state = None
            vnf_instance.save()

        vnf_info['current_error_point'] = EP.NOTIFY_COMPLETED

        notification = vnf_info['notification']
        notification['notificationStatus'] = 'RESULT'
        notification['operationState'] = 'COMPLETED'
        resource_dict = resource_changes.to_dict()
        if resource_dict.get('affected_vnfcs'):
            notification['affectedVnfcs'] = resource_dict.get('affected_vnfcs')
        if resource_dict.get('affected_virtual_links'):
            notification['affectedVirtualLinks'] =\
                resource_dict.get('affected_virtual_links')
        if resource_dict.get('affected_virtual_storages'):
            notification['affectedVirtualStorages'] =\
                resource_dict.get('affected_virtual_storages')
        self.rpc_api.send_notification(context, notification)

    def _scale_resource_update(self, context, vnf_info, vnf_instance,
                               scale_vnf_request,
                               vim_connection_info,
                               error=False):
        vnf_lcm_op_occs = vnf_info['vnf_lcm_op_occ']
        instantiated_vnf_before = \
            copy.deepcopy(vnf_instance.instantiated_vnf_info)

        self._vnf_manager.invoke(
            vim_connection_info.vim_type,
            'scale_resource_update',
            context=context,
            vnf_instance=vnf_instance,
            vnf_info=vnf_info,
            scale_vnf_request=scale_vnf_request,
            vim_connection_info=vim_connection_info
        )
        for scale in vnf_instance.instantiated_vnf_info.scale_status:
            if scale_vnf_request.aspect_id == scale.aspect_id:
                if not error:
                    scale.scale_level = vnf_info['after_scale_level']
                    break
                else:
                    scale.scale_level = vnf_info['scale_level']
                    break
        LOG.debug("vnf_instance.instantiated_vnf_info %s",
        vnf_instance.instantiated_vnf_info)
        affected_vnfcs = []
        affected_virtual_storages = []
        affected_virtual_links = []
        if scale_vnf_request.type == 'SCALE_IN':
            for vnfc in instantiated_vnf_before.vnfc_resource_info:
                vnfc_delete = True
                for rsc in vnf_instance.instantiated_vnf_info.\
                        vnfc_resource_info:
                    if vnfc.compute_resource.resource_id == \
                            rsc.compute_resource.resource_id:
                        vnfc_delete = False
                        break
                if vnfc_delete:
                    affected_vnfc = objects.AffectedVnfc(id=vnfc.id,
                                       vdu_id=vnfc.vdu_id,
                                       change_type='REMOVED',
                                       compute_resource=vnfc.compute_resource)
                    affected_vnfcs.append(affected_vnfc)

            for st in instantiated_vnf_before.virtual_storage_resource_info:
                st_delete = True
                for rsc in vnf_instance.instantiated_vnf_info.\
                        virtual_storage_resource_info:
                    if st.storage_resource.resource_id == \
                            rsc.storage_resource.resource_id:
                        st_delete = False
                        break
                if st_delete:
                    affected_st = objects.AffectedVirtualStorage(
                        id=st.id,
                        virtual_storage_desc_id=st.virtual_storage_desc_id,
                        change_type='REMOVED',
                        storage_resource=st.storage_resource)
                    affected_virtual_storages.append(affected_st)

            for vl in instantiated_vnf_before.vnf_virtual_link_resource_info:
                port_delete = False
                for rsc in vnf_instance.\
                        instantiated_vnf_info.vnf_virtual_link_resource_info:
                    if vl.network_resource.resource_id == \
                            rsc.network_resource.resource_id:
                        if len(vl.vnf_link_ports) != len(rsc.vnf_link_ports):
                            port_delete = True
                            break
                if port_delete:
                    affected_vl = objects.AffectedVirtualLink(
                        id=vl.id,
                        vnf_virtual_link_desc_id=vl.vnf_virtual_link_desc_id,
                        change_type='LINK_PORT_REMOVED',
                        network_resource=vl.network_resource)
                    affected_virtual_links.append(affected_vl)
        else:
            for rsc in vnf_instance.instantiated_vnf_info.vnfc_resource_info:
                vnfc_add = True
                for vnfc in instantiated_vnf_before.vnfc_resource_info:
                    if vnfc.compute_resource.resource_id == \
                            rsc.compute_resource.resource_id:
                        vnfc_add = False
                        break
                if vnfc_add:
                    affected_vnfc = objects.AffectedVnfc(
                        id=rsc.id,
                        vdu_id=rsc.vdu_id,
                        change_type='ADDED',
                        compute_resource=rsc.compute_resource)
                    affected_vnfcs.append(affected_vnfc)
            for rsc in vnf_instance.instantiated_vnf_info.\
                    virtual_storage_resource_info:
                st_add = True
                for st in instantiated_vnf_before.\
                        virtual_storage_resource_info:
                    if st.storage_resource.resource_id == \
                            rsc.storage_resource.resource_id:
                        st_add = False
                        break
                if st_add:
                    affected_st = objects.AffectedVirtualStorage(
                        id=rsc.id,
                        virtual_storage_desc_id=rsc.virtual_storage_desc_id,
                        change_type='ADDED',
                        storage_resource=rsc.storage_resource)
                    affected_virtual_storages.append(affected_st)
            for vl in instantiated_vnf_before.vnf_virtual_link_resource_info:
                port_add = False
                for rsc in vnf_instance.instantiated_vnf_info.\
                        vnf_virtual_link_resource_info:
                    if vl.network_resource.resource_id == \
                            rsc.network_resource.resource_id:
                        if len(vl.vnf_link_ports) != len(rsc.vnf_link_ports):
                            port_add = True
                            break
                if port_add:
                    affected_vl = objects.AffectedVirtualLink(
                        id=vl.id,
                        vnf_virtual_link_desc_id=vl.vnf_virtual_link_desc_id,
                        change_type='LINK_PORT_ADDED',
                        network_resource=vl.network_resource)
                    affected_virtual_links.append(affected_vl)
        resource_changes = objects.ResourceChanges()
        resource_changes.affected_vnfcs = []
        resource_changes.affected_virtual_links = []
        resource_changes.affected_virtual_storages = []
        if 'resource_changes' in \
                vnf_lcm_op_occs and vnf_lcm_op_occs.resource_changes:
            res_chg = vnf_lcm_op_occs.resource_changes
            if 'affected_vnfcs' in res_chg:
                if res_chg.affected_vnfcs and \
                   len(res_chg.affected_vnfcs) > 0:
                    resource_changes.affected_vnfcs.\
                        extend(res_chg.affected_vnfcs)
            if 'affected_virtual_storages' in res_chg:
                if res_chg.affected_virtual_storages and \
                   len(res_chg.affected_virtual_storages) > 0:
                    resource_changes.affected_virtual_storages.extend(
                        res_chg.affected_virtual_storages)
            if 'affected_virtual_links' in res_chg:
                if res_chg.affected_virtual_links and \
                   len(res_chg.affected_virtual_links) > 0:
                    resource_changes.affected_virtual_links.\
                        extend(res_chg.affected_virtual_links)
        resource_changes.affected_vnfcs.extend(affected_vnfcs)
        resource_changes.affected_virtual_storages.extend(
            affected_virtual_storages)
        resource_changes.affected_virtual_links = []
        resource_changes.affected_virtual_links.extend(affected_virtual_links)

        vnf_info['resource_changes'] = resource_changes
        return resource_changes

    @log.log
    @revert_to_error_scale
    def scale_vnf(self, context, vnf_info, vnf_instance, scale_vnf_request):
        LOG.info("Request received for scale vnf '%s'", vnf_instance.id)

        timestamp = datetime.utcnow()
        vnf_lcm_op_occ = vnf_info['vnf_lcm_op_occ']

        vnf_lcm_op_occ.operation_state = 'PROCESSING'
        vnf_lcm_op_occ.state_entered_time = timestamp
        LOG.debug("vnf_lcm_op_occ %s", vnf_lcm_op_occ)
        vnf_lcm_op_occ.save()

        notification = vnf_info['notification']
        notification['operationState'] = 'PROCESSING'
        self.rpc_api.send_notification(context, notification)

        vnf_info['current_error_point'] = EP.NOTIFY_PROCESSING
        vim_info = vnflcm_utils._get_vim(context,
            vnf_instance.vim_connection_info)

        vim_connection_info = objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context)

        scale_id_list, scale_name_list, grp_id = self._scale_vnf_pre(
            context, vnf_info,
            vnf_instance,
            scale_vnf_request,
            vim_connection_info)

        if vnf_info['before_error_point'] <= EP.POST_VIM_CONTROL:
            self._scale_vnf(context, vnf_info, vnf_instance, scale_vnf_request,
                           vim_connection_info, scale_name_list, grp_id,
                           vnf_lcm_op_occ)

            resource_changes = self._scale_resource_update(context, vnf_info,
                                    vnf_instance,
                                    scale_vnf_request,
                                    vim_connection_info)
        else:
            resource_changes = vnf_info.get('resource_changes')
            if not resource_changes:
                resource_changes = self._scale_resource_update(
                    context, vnf_info, vnf_instance, scale_vnf_request,
                    vim_connection_info, error=True)

        vnf_info['current_error_point'] = EP.INTERNAL_PROCESSING

        self._scale_vnf_post(context, vnf_info,
                             vnf_instance,
                             scale_vnf_request,
                             vim_connection_info,
                             scale_id_list,
                             resource_changes)

        LOG.info("Request received for scale vnf '%s' is completed "
                 "successfully", vnf_instance.id)

    def _scale_vnf(self, context, vnf_info, vnf_instance, scale_vnf_request,
            vim_connection_info, scale_name_list, grp_id,
            vnf_lcm_op_occ):
        # action_driver
        LOG.debug("vnf_info['vnfd']['attributes'] %s", (vnf_info
                                                        .get('vnfd', {})
                                                        .get('attributes')))
        self._vnf_manager = driver_manager.DriverManager(
            'tacker.tacker.vnfm.drivers',
            cfg.CONF.tacker.infra_driver)

        if scale_vnf_request.type == 'SCALE_IN':
            action = 'in'
        elif scale_vnf_request.type == 'SCALE_OUT':
            action = 'out'
        else:
            msg = 'Unknown scale type: %s' % scale_vnf_request.type
            raise exceptions.VnfScaleFailed(id=vnf_instance.id, error=msg)

        stack_id = vnf_instance.instantiated_vnf_info.instance_id
        # TODO(h-asahina): change the key name `instance_id` attr to `stack_id`
        policy = {'instance_id': stack_id,
                  'name': scale_vnf_request.aspect_id,
                  'vnf': vnf_info,
                  'action': action}

        LOG.debug(
            "is_reverse: %s",
            scale_vnf_request.additional_params.get('is_reverse'))
        default = None
        if vim_connection_info.vim_type == 'kubernetes':
            policy['vnf_instance_id'] = vnf_lcm_op_occ.get('vnf_instance_id')
            vnf_instance = objects.VnfInstance.get_by_id(context,
                policy['vnf_instance_id'])
            vnfd_dict = vnflcm_utils._get_vnfd_dict(context,
                vnf_instance.vnfd_id,
                vnf_instance.instantiated_vnf_info.flavour_id)
            tosca = tosca_template.ToscaTemplate(
                parsed_params={}, a_file=False, yaml_dict_tpl=vnfd_dict)
            extract_policy_infos = vnflcm_utils.get_extract_policy_infos(tosca)
            policy['vdu_defs'] = vnflcm_utils.get_target_vdu_def_dict(
                extract_policy_infos=extract_policy_infos,
                aspect_id=scale_vnf_request.aspect_id,
                tosca=tosca)
            policy['delta_num'] = vnflcm_utils.get_scale_delta_num(
                extract_policy_infos=extract_policy_infos,
                aspect_id=scale_vnf_request.aspect_id)
        elif vim_connection_info.vim_type == 'openstack':
            # NOTE(ueha): The logic of Scale for OpenStack VIM is widely hard
            # coded with `vnf_info`. This dependency is to be refactored in
            # future.
            scale_json = vnf_info['attributes']['scale_group']
            scale_group_dict = jsonutils.loads(scale_json)
            key_aspect = scale_vnf_request.aspect_id
            default = scale_group_dict['scaleGroupDict'][key_aspect]['default']
        else:
            msg = 'Unknown vim type: %s' % vim_connection_info.vim_type
            raise exceptions.VnfScaleFailed(id=vnf_instance.id, error=msg)

        if (scale_vnf_request.type == 'SCALE_IN' and
                scale_vnf_request.additional_params['is_reverse'] == 'True'):
            self._vnf_manager.invoke(
                vim_connection_info.vim_type,
                'scale_in_reverse',
                plugin=self,
                context=context,
                auth_attr=vim_connection_info.access_info,
                vnf_info=vnf_info,
                scale_vnf_request=scale_vnf_request,
                region_name=vim_connection_info.access_info.get('region_name'),
                scale_name_list=scale_name_list,
                grp_id=grp_id,
                vnf_instance=vnf_instance
            )
            vnf_info['current_error_point'] = EP.POST_VIM_CONTROL
            self._vnf_manager.invoke(
                vim_connection_info.vim_type,
                'scale_update_wait',
                plugin=self,
                context=context,
                auth_attr=vim_connection_info.access_info,
                vnf_info=vnf_info,
                region_name=vim_connection_info.access_info.get('region_name')
            )
        elif scale_vnf_request.type == 'SCALE_OUT' and default == 0:
            self._vnf_manager.invoke(
                vim_connection_info.vim_type,
                'scale_out_initial',
                plugin=self,
                context=context,
                auth_attr=vim_connection_info.access_info,
                vnf_info=vnf_info,
                scale_vnf_request=scale_vnf_request,
                region_name=vim_connection_info.access_info.get('region_name'),
                vnf_instance=vnf_instance
            )
            vnf_info['current_error_point'] = EP.POST_VIM_CONTROL
            self._vnf_manager.invoke(
                vim_connection_info.vim_type,
                'scale_update_wait',
                plugin=self,
                context=context,
                auth_attr=vim_connection_info.access_info,
                vnf_info=vnf_info,
                region_name=vim_connection_info.access_info.get('region_name')
            )
        else:
            for _ in range(scale_vnf_request.number_of_steps):
                last_event_id = self._vnf_manager.invoke(
                    vim_connection_info.vim_type,
                    'scale',
                    plugin=self,
                    context=context,
                    auth_attr=vim_connection_info.access_info,
                    policy=policy,
                    region_name=vim_connection_info.access_info.get('\
                        region_name')
                )
                vnf_info['current_error_point'] = EP.POST_VIM_CONTROL
                self._vnf_manager.invoke(
                    vim_connection_info.vim_type,
                    'scale_wait',
                    plugin=self,
                    context=context,
                    auth_attr=vim_connection_info.access_info,
                    policy=policy,
                    region_name=vim_connection_info.access_info.get('\
                        region_name'),
                    last_event_id=last_event_id)

    def _term_resource_update(self, context, vnf_info, vnf_instance,
                              error=False):
        if not vnf_instance.instantiated_vnf_info:
            resource_changes = objects.ResourceChanges()
            resource_changes.affected_vnfcs = []
            resource_changes.affected_virtual_links = []
            resource_changes.affected_virtual_storages = []
            vnf_info['resource_changes'] = resource_changes
            return resource_changes
        instantiated_vnf_before = copy.deepcopy(
            vnf_instance.instantiated_vnf_info)
        vnf_instance.instantiated_vnf_info.reinitialize()
        if not error:
            vnf_instance.vim_connection_info = []
            vnf_instance.task_state = None
        LOG.debug(
            "vnf_instance.instantiated_vnf_info %s",
            vnf_instance.instantiated_vnf_info)
        affected_vnfcs = []
        affected_virtual_storages = []
        affected_virtual_links = []
        for vnfc in instantiated_vnf_before.vnfc_resource_info:
            vnfc_delete = True
            for rsc in vnf_instance.instantiated_vnf_info.vnfc_resource_info:
                if vnfc.compute_resource.resource_id == \
                        rsc.compute_resource.resource_id:
                    vnfc_delete = False
                    break
            if vnfc_delete:
                affected_vnfc = objects.AffectedVnfc(
                    id=vnfc.id,
                    vdu_id=vnfc.vdu_id,
                    change_type='REMOVED',
                    compute_resource=vnfc.compute_resource)
                affected_vnfcs.append(affected_vnfc)

        for st in instantiated_vnf_before.virtual_storage_resource_info:
            st_delete = True
            for rsc in \
                    vnf_instance.instantiated_vnf_info.\
                    virtual_storage_resource_info:
                if st.storage_resource.resource_id == \
                        rsc.storage_resource.resource_id:
                    st_delete = False
                    break
            if st_delete:
                affected_st = objects.AffectedVirtualStorage(
                    id=st.id,
                    virtual_storage_desc_id=st.virtual_storage_desc_id,
                    change_type='REMOVED',
                    storage_resource=st.storage_resource)
                affected_virtual_storages.append(affected_st)

        for vl in instantiated_vnf_before.vnf_virtual_link_resource_info:
            vm_delete = False
            for rsc in \
                    vnf_instance.instantiated_vnf_info.\
                    vnf_virtual_link_resource_info:
                if st.network_resource.resource_id == \
                        rsc.network_resource.resource_id:
                    vm_delete = False
                    break
            if vm_delete:
                affected_vl = objects.AffectedVirtualLink(
                    id=vl.id,
                    vnf_virtual_link_desc_id=vl.vnf_virtual_link_desc_id,
                    change_type='REMOVED',
                    network_resource=vl.network_resource)
                affected_virtual_links.append(affected_vl)
        vnf_lcm_op_occs = vnf_info['vnf_lcm_op_occ']
        resource_changes = objects.ResourceChanges()
        resource_changes.affected_vnfcs = []
        resource_changes.affected_virtual_links = []
        resource_changes.affected_virtual_storages = []
        if 'resource_changes' in vnf_lcm_op_occs and \
                vnf_lcm_op_occs.resource_changes:
            res_chg = vnf_lcm_op_occs.resource_changes
            if 'affected_vnfcs' in res_chg:
                if res_chg.affected_vnfcs and \
                        len(res_chg.affected_vnfcs) > 0:
                    resource_changes.affected_vnfcs.extend(
                        res_chg.affected_vnfcs)
            if 'affected_virtual_storages' in res_chg:
                if res_chg.affected_virtual_storages and \
                        len(res_chg.affected_virtual_storages) > 0:
                    resource_changes.affected_virtual_storages.extend(
                        res_chg.affected_virtual_storages)
            if 'affected_virtual_links' in res_chg:
                if res_chg.affected_virtual_links and \
                        len(res_chg.affected_virtual_links) > 0:
                    resource_changes.affected_virtual_links.extend(
                        res_chg.affected_virtual_links)
        resource_changes.affected_vnfcs.extend(affected_vnfcs)
        resource_changes.affected_virtual_storages.extend(
            affected_virtual_storages)
        resource_changes.affected_virtual_links.extend(affected_virtual_links)

        vnf_info['resource_changes'] = resource_changes
        return resource_changes

    def _rollback_vnf_pre(
            self,
            context,
            vnf_info,
            vnf_instance,
            operation_params,
            vim_connection_info):
        vnf_lcm_op_occs = vnf_info['vnf_lcm_op_occ']
        scale_id_list = []
        scale_name_list = []
        grp_id = None
        self._update_vnf_rollback_pre(context, vnf_info)
        if vnf_lcm_op_occs.operation == 'SCALE':
            if vim_connection_info.vim_type != 'kubernetes':
                # NOTE(ueha): The logic of Scale for OpenStack VIM is widely
                # hard coded with `vnf_info`. This dependency is to be
                # refactored in future.
                scaleGroupDict = jsonutils.loads(
                    vnf_info['attributes']['scale_group'])
                cap_size = scaleGroupDict['scaleGroupDict'][operation_params
                    ['aspect_id']]['default']
                vnf_info['res_num'] = cap_size
            scale_vnf_request = objects.ScaleVnfRequest.obj_from_primitive(
                operation_params, context=context)
            for scale in vnf_instance.instantiated_vnf_info.scale_status:
                if scale_vnf_request.aspect_id == scale.aspect_id:
                    vnf_info['after_scale_level'] = scale.scale_level
                    break
        if vnf_lcm_op_occs.operation == 'SCALE' \
           and vnf_lcm_op_occs.error_point >= EP.POST_VIM_CONTROL:
            scale_id_list, scale_name_list, grp_id = self._vnf_manager.invoke(
                vim_connection_info.vim_type,
                'get_rollback_ids',
                plugin=self,
                context=context,
                vnf_dict=vnf_info,
                aspect_id=operation_params['aspect_id'],
                auth_attr=vim_connection_info.access_info,
                region_name=vim_connection_info.access_info.get('region_name')
            )
        if vnf_lcm_op_occs.error_point == EP.NOTIFY_COMPLETED:
            if vnf_lcm_op_occs.operation == 'SCALE':
                vnfd_dict = vnflcm_utils._get_vnfd_dict(
                    context, vnf_instance.vnfd_id,
                    vnf_instance.instantiated_vnf_info.flavour_id)
                vnf_info['action'] = 'in'
                if len(scale_id_list) != 0:
                    kwargs = {'scale_name_list': scale_name_list}
                    # TODO(LiangLu): grant_request here is planned to pass
                    # as a parameter, however due to grant_request are not
                    # passed from conductor to vnflcm_driver, thus we put Null
                    # value to grant and grant_reqeust temporary.
                    # This part will be updated in next release.
                    self._mgmt_manager.invoke(
                        self._load_vnf_interface(context, 'scale_start',
                                                 vnf_instance, vnfd_dict),
                        'scale_start', context=context,
                        vnf_instance=vnf_instance,
                        scale_vnf_request=scale_vnf_request,
                        grant=vnf_info.get('grant'), grant_request=None,
                        **kwargs)

            else:
                vnfd_dict = vnflcm_utils._get_vnfd_dict(
                    context, vnf_instance.vnfd_id,
                    vnf_instance.instantiated_vnf_info.flavour_id)
                # TODO(LiangLu): grant_request and grant here is planned to
                # pass as a parameter, however due to they are not
                # passed from conductor to vnflcm_driver, thus we put Null
                # value to grant and grant_reqeust temporary.
                # This part will be updated in next release.
                if len(scale_id_list) != 0:
                    self._mgmt_manager.invoke(
                        self._load_vnf_interface(
                            context, 'terminate_start',
                            vnf_instance, vnfd_dict),
                        'terminate_start', context=context,
                        vnf_instance=vnf_instance,
                        terminate_vnf_request=None,
                        grant=None, grant_request=None)
            vnf_lcm_op_occs.error_point = EP.VNF_CONFIG_END

        return scale_name_list, grp_id

    def _rollback_vnf(
            self,
            context,
            vnf_info,
            vnf_instance,
            operation_params,
            vim_connection_info,
            scale_name_list,
            grp_id):
        vnf_lcm_op_occs = vnf_info['vnf_lcm_op_occ']
        if vnf_lcm_op_occs.error_point >= EP.POST_VIM_CONTROL:
            if vnf_lcm_op_occs.operation == 'SCALE':
                scale_vnf_request = objects.ScaleVnfRequest.obj_from_primitive(
                    operation_params, context=context)
                self._vnf_manager.invoke(
                    vim_connection_info.vim_type,
                    'scale_in_reverse',
                    plugin=self,
                    context=context,
                    auth_attr=vim_connection_info.access_info,
                    vnf_info=vnf_info,
                    scale_vnf_request=scale_vnf_request,
                    region_name=vim_connection_info.access_info.get(
                        'region_name'),
                    scale_name_list=scale_name_list,
                    grp_id=grp_id,
                    vnf_instance=vnf_instance)
                self._vnf_manager.invoke(
                    vim_connection_info.vim_type,
                    'scale_update_wait',
                    plugin=self,
                    context=context,
                    auth_attr=vim_connection_info.access_info,
                    vnf_info=vnf_info,
                    region_name=vim_connection_info.access_info.get(
                        'region_name'))

            else:
                instance_id = vnf_instance.instantiated_vnf_info.instance_id
                access_info = vim_connection_info.access_info
                self._vnf_manager.invoke(vim_connection_info.vim_type,
                    'delete', plugin=self, context=context,
                    vnf_id=instance_id, auth_attr=access_info,
                    vnf_instance=vnf_instance)

                self._vnf_manager.invoke(vim_connection_info.vim_type,
                    'delete_wait', plugin=self, context=context,
                    vnf_id=instance_id, auth_attr=access_info,
                    vnf_instance=vnf_instance)

        vnf_lcm_op_occs.error_point = EP.PRE_VIM_CONTROL

    def _update_vnf_rollback_pre(self, context, vnf_info):
        self._vnfm_plugin._update_vnf_rollback_pre(context, vnf_info)

    def _update_vnf_rollback(self, context, vnf_info,
                             vnf_instance, vnf_lcm_op_occs):
        if vnf_lcm_op_occs.operation == 'SCALE':
            status = 'ACTIVE'
        else:
            status = 'INACTIVE'
        vnf_instance.task_state = None
        self._vnfm_plugin._update_vnf_rollback(context, vnf_info, 'ERROR',
                                               status)
        vnf_lcm_op_occs.state_entered_time = timeutils.utcnow()
        vnf_lcm_op_occs.save()
        vnf_instance.save()

    def _update_vnf_rollback_status_err(self, context, vnf_info):
        self._vnfm_plugin.update_vnf_rollback_status_err(context, vnf_info)

    def _rollback_vnf_post(
            self,
            context,
            vnf_info,
            vnf_instance,
            operation_params,
            vim_connection_info):
        vnf_lcm_op_occs = vnf_info['vnf_lcm_op_occ']
        if vnf_lcm_op_occs.operation == 'SCALE':
            scale_vnf_request = objects.ScaleVnfRequest.obj_from_primitive(
                operation_params, context=context)
            scale_vnf_request_copy = copy.deepcopy(scale_vnf_request)
            scale_vnf_request_copy.type = 'SCALE_IN'
            resource_changes = self._scale_resource_update(context, vnf_info,
                                    vnf_instance,
                                    scale_vnf_request_copy,
                                    vim_connection_info)

        else:
            vnfd_dict = vnflcm_utils.get_vnfd_dict(
                context, vnf_instance.vnfd_id,
                vnf_instance.instantiated_vnf_info.flavour_id)
            # TODO(LiangLu): grant_request and grant here is planned to
            # pass as a parameter, however due to they are not
            # passed from conductor to vnflcm_driver, thus we put Null
            # value to grant and grant_reqeust temporary.
            # This part will be updated in next release.
            self._mgmt_manager.invoke(
                self._load_vnf_interface(
                    context, 'terminate_end',
                    vnf_instance, vnfd_dict),
                'terminate_end', context=context,
                vnf_instance=vnf_instance,
                terminate_vnf_request=None,
                grant=None, grant_request=None)
            resource_changes = self._term_resource_update(
                context, vnf_info, vnf_instance)

        vnf_lcm_op_occs.error_point = EP.VNF_CONFIG_START

        timestamp = datetime.utcnow()
        vnf_lcm_op_occs.operation_state = 'ROLLED_BACK'
        vnf_lcm_op_occs.state_entered_time = timestamp
        vnf_lcm_op_occs.resource_changes = resource_changes
        self._update_vnf_rollback(context, vnf_info,
                                 vnf_instance,
                                 vnf_lcm_op_occs)
        notification = vnf_info['notification']
        notification['notificationStatus'] = 'RESULT'
        notification['operationState'] = 'ROLLED_BACK'
        resource_dict = resource_changes.to_dict()
        if resource_dict.get('affected_vnfcs'):
            notification['affectedVnfcs'] = resource_dict.get('affected_vnfcs')
        if resource_dict.get('affected_virtual_links'):
            notification['affectedVirtualLinks'] = \
                resource_dict.get('affected_virtual_links')
        if resource_dict.get('affected_virtual_storages'):
            notification['affectedVirtualStorages'] = \
                resource_dict.get('affected_virtual_storages')
        self.rpc_api.send_notification(context, notification)

    @log.log
    @revert_to_error_rollback
    def rollback_vnf(self, context, vnf_info, vnf_instance, operation_params):
        LOG.info("Request received for rollback vnf '%s'", vnf_instance.id)
        vnf_lcm_op_occs = vnf_info['vnf_lcm_op_occ']
        if vnf_lcm_op_occs.operation == 'SCALE':
            scale_vnf_request = objects.ScaleVnfRequest.obj_from_primitive(
                operation_params, context=context)
            for scale in vnf_instance.instantiated_vnf_info.scale_status:
                if scale_vnf_request.aspect_id == scale.aspect_id:
                    vnf_info['after_scale_level'] = scale.scale_level
                    break

        timestamp = datetime.utcnow()

        vnf_lcm_op_occs.operation_state = 'ROLLING_BACK'
        vnf_lcm_op_occs.state_entered_time = timestamp
        LOG.debug("vnf_lcm_op_occs %s", vnf_lcm_op_occs)

        insta_url = CONF.vnf_lcm.endpoint_url + \
            "/vnflcm/v1/vnf_instances/" + \
            vnf_instance.id
        vnflcm_url = CONF.vnf_lcm.endpoint_url + \
            "/vnflcm/v1/vnf_lcm_op_occs/" + \
            vnf_lcm_op_occs.id
        notification = {}
        notification['notificationType'] = \
            'VnfLcmOperationOccurrenceNotification'
        notification['vnfInstanceId'] = vnf_instance.id
        notification['notificationStatus'] = 'START'
        notification['operation'] = vnf_lcm_op_occs.operation
        notification['operationState'] = 'ROLLING_BACK'
        if vnf_lcm_op_occs.operation == 'SCALE':
            notification['isAutomaticInvocation'] = \
                vnf_lcm_op_occs.is_automatic_invocation
        else:
            notification['isAutomaticInvocation'] = False
        notification['vnfLcmOpOccId'] = vnf_lcm_op_occs.id
        notification['_links'] = {}
        notification['_links']['vnfInstance'] = {}
        notification['_links']['vnfInstance']['href'] = insta_url
        notification['_links']['vnfLcmOpOcc'] = {}
        notification['_links']['vnfLcmOpOcc']['href'] = vnflcm_url
        vnf_info['notification'] = notification
        vnf_lcm_op_occs.save()
        self.rpc_api.send_notification(context, notification)

        vim_info = vnflcm_utils._get_vim(context,
            vnf_instance.vim_connection_info)

        vim_connection_info = objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context)

        scale_name_list, grp_id = self._rollback_vnf_pre(
            context, vnf_info, vnf_instance,
            operation_params, vim_connection_info)

        self._rollback_vnf(
            context,
            vnf_info,
            vnf_instance,
            operation_params,
            vim_connection_info,
            scale_name_list,
            grp_id)

        self._rollback_vnf_post(
            context,
            vnf_info,
            vnf_instance,
            operation_params,
            vim_connection_info)

    def _change_ext_conn_vnf(self, context, vnf_instance, vnf_dict,
            vim_connection_info, change_ext_conn_req):
        inst_vnf_info = vnf_instance.instantiated_vnf_info
        try:
            self._vnf_manager.invoke(
                vim_connection_info.vim_type, 'change_ext_conn_vnf',
                context=context, vnf_instance=vnf_instance, vnf_dict=vnf_dict,
                vim_connection_info=vim_connection_info,
                change_ext_conn_req=change_ext_conn_req)
        except Exception as exp:
            with excutils.save_and_reraise_exception() as exc_ctxt:
                exc_ctxt.reraise = False
                LOG.error("Failed to change external connectivity "
                          "vnf %(id)s in infra driver. "
                          "Error: %(error)s", {"id": vnf_instance.id, "error":
                          encodeutils.exception_to_unicode(exp)})
                raise exceptions.VnfChangeExtConnFailed(id=vnf_instance.id,
                    error=encodeutils.exception_to_unicode(exp))
        vnf_dict['current_error_point'] = fields.ErrorPoint.POST_VIM_CONTROL
        try:
            self._vnf_manager.invoke(
                vim_connection_info.vim_type, 'change_ext_conn_vnf_wait',
                context=context, vnf_instance=vnf_instance,
                vim_connection_info=vim_connection_info)
        except Exception as exp:
            LOG.error("Failed to update vnf %(id)s resources for instance "
                      "%(instance)s. Error: %(error)s",
                      {'id': vnf_instance.id, 'instance':
                      inst_vnf_info.instance_id, 'error':
                      encodeutils.exception_to_unicode(exp)})
            raise exceptions.VnfChangeExtConnWaitFailed(
                id=vnf_instance.id,
                error=encodeutils.exception_to_unicode(exp))

    @log.log
    @revert_to_error_task_state
    def change_ext_conn_vnf(
            self,
            context,
            vnf_instance,
            vnf_dict,
            change_ext_conn_req):
        LOG.info("Request received for changing external connectivity "
                 "vnf '%s'", vnf_instance.id)

        vnfd_dict = vnflcm_utils._get_vnfd_dict(
            context, vnf_instance.vnfd_id,
            vnf_instance.instantiated_vnf_info.flavour_id)

        vnf_dict['current_error_point'] = EP.VNF_CONFIG_START
        if vnf_dict['before_error_point'] <= EP.VNF_CONFIG_START:
            # TODO(esto-aln): grant_request here is planned to pass
            # as a parameter, however due to grant_request are not
            # passed from conductor to vnflcm_driver, thus we put Null
            # value to grant and grant_reqeust temporary.
            # This part will be updated in next release.
            self._mgmt_manager.invoke(
                self._load_vnf_interface(
                    context, 'change_external_connectivity_start',
                    vnf_instance, vnfd_dict),
                'change_external_connectivity_start', context=context,
                vnf_instance=vnf_instance,
                change_ext_conn_request=change_ext_conn_req,
                grant=vnf_dict.get('grant'), grant_request=None)

        vnf_dict['current_error_point'] = EP.PRE_VIM_CONTROL
        if vnf_dict['before_error_point'] <= EP.POST_VIM_CONTROL:
            vim_info = vnflcm_utils._get_vim(context,
                vnf_instance.vim_connection_info)

            vim_connection_info = \
                objects.VimConnectionInfo.obj_from_primitive(
                    vim_info, context)

            self._change_ext_conn_vnf(context, vnf_instance, vnf_dict,
                vim_connection_info, change_ext_conn_req)

        # Since there is no processing corresponding to
        # EP.INTERNAL_PROCESSING, it transitions to EP.VNF_CONFIG_END.
        vnf_dict['current_error_point'] = EP.VNF_CONFIG_END
        if vnf_dict['before_error_point'] <= EP.VNF_CONFIG_END:
            # TODO(esto-aln): grant_request here is planned to pass
            # as a parameter, however due to grant_request are not
            # passed from conductor to vnflcm_driver, thus we put Null
            # value to grant and grant_reqeust temporary.
            # This part will be updated in next release.
            self._mgmt_manager.invoke(
                self._load_vnf_interface(
                    context, 'change_external_connectivity_end',
                    vnf_instance, vnfd_dict),
                'change_external_connectivity_end', context=context,
                vnf_instance=vnf_instance,
                change_ext_conn_request=change_ext_conn_req,
                grant=vnf_dict.get('grant'), grant_request=None)

        LOG.info("Request received for changing external connectivity "
                 "vnf '%s' is completed successfully", vnf_instance.id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet
import os
import paramiko
import time

from oslo_log import log as logging

from tacker.common import cmd_executer
from tacker.common import exceptions
from tacker.common import log
from tacker import objects
from tacker.vnflcm import utils as vnflcm_utils
from tacker.vnfm.infra_drivers.openstack import heat_client as hc
from tacker.vnfm.mgmt_drivers import vnflcm_abstract_driver

LOG = logging.getLogger(__name__)
FREE5GC_CMD_TIMEOUT = 30
FREE5GC_MODIFY_CONFIG_TIMEOUT = 60


class Free5gcMgmtDriver(vnflcm_abstract_driver.VnflcmMgmtAbstractDriver):
    def get_type(self):
        return 'mgmt-drivers-free5gc'

    def get_name(self):
        return 'mgmt-drivers-free5gc'

    def get_description(self):
        return 'Tacker VNFMgmt Free5gc Driver'

    @log.log
    def instantiate_start(self, context, vnf_instance,
                          instantiate_vnf_request, grant,
                          grant_request, **kwargs):
        pass

    def _get_vim_connection_info(self, context, instantiate_vnf_req):

        vim_info = vnflcm_utils._get_vim(context,
                instantiate_vnf_req.vim_connection_info)

        vim_connection_info = objects.VimConnectionInfo.obj_from_primitive(
            vim_info, context)

        return vim_connection_info

    def _check_values(self, additional_param):
        if not additional_param.get('username'):
            LOG.error('The username in the '
                      'additionalParams is invalid.')
            raise exceptions.MgmtDriverNotFound(param='username')
        if not additional_param.get('password'):
            LOG.error('The password in the '
                      'additionalParams is invalid.')
            raise exceptions.MgmtDriverNotFound(param='password')
        if not additional_param.get('aspect_id'):
            LOG.error('The aspect_id in the '
                      'additionalParams is invalid.')
            raise exceptions.MgmtDriverNotFound(param='password')
        if not additional_param.get('ssh_cp_name'):
            LOG.error('The ssh_cp_name in the '
                      'additionalParams is invalid.')
            raise exceptions.MgmtDriverNotFound(param='password')
        if not additional_param.get('amf_cp_name'):
            LOG.error('The amf_cp_name in the '
                      'additionalParams is invalid.')
            raise exceptions.MgmtDriverNotFound(param='password')
        if not additional_param.get('smf_cp_name'):
            LOG.error('The smf_cp_name in the '
                      'additionalParams is invalid.')
            raise exceptions.MgmtDriverNotFound(param='password')
        if not additional_param.get('upf_cp_name'):
            LOG.error('The upf_cp_name in the '
                      'additionalParams is invalid.')
            raise exceptions.MgmtDriverNotFound(param='password')
        if not additional_param.get('modify_script_path'):
            LOG.error('The modify_script_path in the '
                      'additionalParams is invalid.')
            raise exceptions.MgmtDriverNotFound(param='ssh_cp_name')

    def _get_group_resources_list(
            self, heatclient, stack_id, aspect_id, additional_params):
        nest_resources_list = heatclient.resources.list(stack_id=stack_id)
        group_stack_name = aspect_id
        if 'lcm-operation-user-data' in additional_params.keys() and \
                'lcm-operation-user-data-class' in additional_params.keys():
            group_stack_name = group_stack_name + '_group'
        group_stack_id = ""
        for nest_resources in nest_resources_list:
            if nest_resources.resource_name == group_stack_name:
                group_stack_id = nest_resources.physical_resource_id
        if not group_stack_id:
            LOG.error('No stack id matching the group was found.')
            raise exceptions.MgmtDriverOtherError(
                error_message="No stack id matching the group was found")
        group_resources_list = heatclient.resources.list(
            stack_id=group_stack_id)
        return group_resources_list

    def _init_commander_and_send_modify_scripts(self, user, password, host,
                        vnf_package_path=None, script_path=None):
        retry = 4
        while retry > 0:
            try:
                if vnf_package_path and script_path:
                    connect = paramiko.Transport(host, 22)
                    connect.connect(username=user, password=password)
                    sftp = paramiko.SFTPClient.from_transport(connect)
                    # put script file content to '/tmp/modify_config.sh'
                    sftp.put(os.path.join(vnf_package_path, script_path),
                             "/tmp/modify_config.sh")
                    connect.close()
                commander = cmd_executer.RemoteCommandExecutor(
                    user=user, password=password, host=host,
                    timeout=30)
                return commander
            except paramiko.SSHException as e:
                LOG.debug(e)
                retry -= 1
                if retry == 0:
                    LOG.error(e)
                    raise paramiko.SSHException()
                time.sleep(120)

    def _execute_command(self, commander, ssh_command, timeout, type, retry,
            input_data=None):
        eventlet.monkey_patch()
        while retry >= 0:
            try:
                with eventlet.Timeout(timeout, True):
                    result = commander.execute_command(
                        ssh_command, input_data=input_data)
                    break
            except eventlet.timeout.Timeout:
                LOG.debug('It is time out, When execute command: '
                          '{}.'.format(ssh_command))
                retry -= 1
                if retry < 0:
                    LOG.error('It is time out, When execute command: '
                              '{}.'.format(ssh_command))
                    raise exceptions.MgmtDriverOtherError(
                        error_message='It is time out, When execute command: '
                                      '{}.'.format(ssh_command))
                time.sleep(30)
        if type == 'common':
            if result.get_return_code() != 0:
                err = result.get_stderr()
                LOG.error(err)
                raise exceptions.MgmtDriverRemoteCommandError(err_info=err)
        return result.get_stdout()

    def _get_smf_and_amf_ip(
            self, heatclient, stack_id, amf_cp_name, smf_cp_name):
        amf_cplan_ip = heatclient.resources.get(
            stack_id=stack_id,
            resource_name=amf_cp_name).attributes.get(
            'fixed_ips')[0].get('ip_address')
        smf_pfcp_ip = heatclient.resources.get(
            stack_id=stack_id,
            resource_name=smf_cp_name).attributes.get(
            'fixed_ips')[0].get('ip_address')
        return amf_cplan_ip, smf_pfcp_ip

    def _modify_config_path_and_start_process(
            self, group_resources_list, heatclient,
            vnf_package_path, additional_param, operation_type,
            **kwargs):
        ssh_cp_name = additional_param.get('ssh_cp_name')
        username = additional_param.get('username')
        password = additional_param.get('password')
        amf_cp_name = additional_param.get('amf_cp_name')
        smf_cp_name = additional_param.get('smf_cp_name')
        upf_cp_name = additional_param.get('upf_cp_name')
        modify_script_path = additional_param.get('modify_script_path')

        if operation_type == 'HEAL':
            target_physical_resource_ids = \
                kwargs['target_physical_resource_ids']
            vnf_ssh_ip, heal_stack_id = self._get_vnf_ssh_ip(
                group_resources_list, heatclient, ssh_cp_name,
                target_physical_resource_ids, upf_cp_name)
            ssh_ip = vnf_ssh_ip['heal']
            amf_cplan_ip, smf_pfcp_ip = \
                self._get_smf_and_amf_ip(
                    heatclient, heal_stack_id, amf_cp_name,
                    smf_cp_name)
            upf_pfcp_ip_master = vnf_ssh_ip['upf_pfcp_ip']

        for group_resource in group_resources_list:
            stack_id = group_resource.physical_resource_id
            resource_name = ssh_cp_name
            resource_info = heatclient.resources.get(
                stack_id=stack_id,
                resource_name=resource_name)
            # get ssh_ip
            upf_pfcp_ip_list = []
            if operation_type == 'INSTANTIATE':
                if resource_info.attributes.get('floating_ip_address'):
                    ssh_ip = resource_info.attributes.get(
                        'floating_ip_address')
                else:
                    ssh_ip = heatclient.resources.get(
                        stack_id=stack_id,
                        resource_name=resource_name).attributes.get(
                        'fixed_ips')[0].get('ip_address')
                # get amf,smf,upf's config ip
                amf_cplan_ip, smf_pfcp_ip = \
                    self._get_smf_and_amf_ip(
                        heatclient, stack_id, amf_cp_name,
                        smf_cp_name)
                upf_pfcp_ip_master = heatclient.resources.get(
                    stack_id=stack_id,
                    resource_name=upf_cp_name).attributes.get(
                    'fixed_ips')[0].get('ip_address')
                upf_pfcp_ip_list.append(upf_pfcp_ip_master)
            else:
                upf_pfcp_ip = heatclient.resources.get(
                    stack_id=stack_id,
                    resource_name=upf_cp_name).attributes.get(
                    'fixed_ips')[0].get('ip_address')
                upf_pfcp_ip_list.append(upf_pfcp_ip)

        upf_pfcp_ips_str = ','.join(upf_pfcp_ip_list)
        # get all config file path
        amf_config_path = additional_param.get('amf_config_path')
        if not amf_config_path:
            amf_config_path = '~/free5gc/config/amfcfg.conf'
        smf_config_path = additional_param.get('smf_config_path')
        if not smf_config_path:
            smf_config_path = '~/free5gc/config/smfcfg.conf'
        upf_config_path = additional_param.get('upf_config_path')
        if not upf_config_path:
            upf_config_path = '~/free5gc/src/upf/build/config/upfcfg.yaml'
        start_process_script_path = \
            additional_param.get('start_process_script_path')
        if not start_process_script_path:
            start_process_script_path = '~/free5gc/run.sh'
            start_process_script_dir = \
                start_process_script_path.replace('/run.sh', '')
        web_console_script_path = \
            additional_param.get('web_console_script_path')
        if not web_console_script_path:
            web_console_script_path = '~/free5gc/webconsole/server.go'
        web_console_script_path = web_console_script_path.replace('/server.go',
                                                                  '')
        commander = self._init_commander_and_send_modify_scripts(
            username, password, ssh_ip,
            vnf_package_path, modify_script_path)
        # modify config file
        ssh_command = "bash /tmp/modify_config.sh " \
                      "-A {amf_config_path} -a {amf_cplan_ip} " \
                      "-S {smf_config_path} -s {smf_pfcp_ip} " \
                      "-U {upf_config_path} -u {upf_pfcp_ip} " \
                      "-l {upf_pfcp_ip_str}".format(
                          amf_config_path=amf_config_path,
                          smf_config_path=smf_config_path,
                          upf_config_path=upf_config_path,
                          amf_cplan_ip=amf_cplan_ip,
                          smf_pfcp_ip=smf_pfcp_ip,
                          upf_pfcp_ip=upf_pfcp_ip_master,
                          upf_pfcp_ip_str=upf_pfcp_ips_str)
        self._execute_command(
            commander, ssh_command, FREE5GC_MODIFY_CONFIG_TIMEOUT,
            'common', 0, input_data=password)
        # start all process
        ssh_command = ("cd {start_process_script_dir};"
                       "touch result.txt;"
                       "nohup ./run.sh > result.txt 2>&1 &".format(
                           start_process_script_dir=start_process_script_dir))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT, 'common', 0)
        time.sleep(120)
        # confirm process running
        ssh_command = r'ps -ef | grep "\./bin"'
        results = self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT, 'common', 0)
        free5gc_list = ['free5gc-upfd', 'nrf', 'amf',
                        'smf', 'udr', 'pcf', 'udm', 'nssf', 'ausf']
        count = 0
        for result in results:
            for nf in free5gc_list:
                if nf in result:
                    count = count + 1
                    break
        if count != 10:
            LOG.error('The process of free5gc start failed.')
            raise exceptions.MgmtDriverOtherError(
                error_message='The process of free5gc start failed.')
        # start web console
        ssh_command = ("cd {web_console_script_path}; touch webconsole.txt; "
                       "nohup go run server.go > webconsole.txt 2>&1 &".format(
                           web_console_script_path=web_console_script_path))
        self._execute_command(
            commander, ssh_command, FREE5GC_CMD_TIMEOUT, 'common', 0)
        commander.close_session()

    @log.log
    def instantiate_end(self, context, vnf_instance,
                        instantiate_vnf_request, grant,
                        grant_request, **kwargs):
        # get vim_connect_info
        if hasattr(instantiate_vnf_request, 'vim_connection_info'):
            vim_connection_info = self._get_vim_connection_info(
                context, instantiate_vnf_request)
        else:
            vim_connection_info = self._get_vim_connection_info(
                context, vnf_instance)
        additional_param = instantiate_vnf_request.additional_params.get(
            'free5gc', {})
        self._check_values(additional_param)
        aspect_id = additional_param.get('aspect_id')

        vnf_package_path = vnflcm_utils._get_vnf_package_path(
            context, vnf_instance.vnfd_id)

        access_info = vim_connection_info.access_info
        heatclient = hc.HeatClient(access_info)
        nest_stack_id = vnf_instance.instantiated_vnf_info.instance_id
        group_resources_list = self._get_group_resources_list(
            heatclient, nest_stack_id, aspect_id,
            instantiate_vnf_request.additional_params)
        self._modify_config_path_and_start_process(
            group_resources_list, heatclient, vnf_package_path,
            additional_param, 'INSTANTIATE')

    @log.log
    def terminate_start(self, context, vnf_instance,
                        terminate_vnf_request, grant,
                        grant_request, **kwargs):
        pass

    @log.log
    def terminate_end(self, context, vnf_instance,
                      terminate_vnf_request, grant,
                      grant_request, **kwargs):
        pass

    @log.log
    def scale_start(self, context, vnf_instance,
                    scale_vnf_request, grant,
                    grant_request, **kwargs):
        pass

    @log.log
    def scale_end(self, context, vnf_instance,
                  scale_vnf_request, grant,
                  grant_request, **kwargs):
        additional_param = vnf_instance.instantiated_vnf_info. \
            additional_params.get('free5gc', {})
        ssh_cp_name = additional_param.get('ssh_cp_name')
        username = additional_param.get('username')
        password = additional_param.get('password')
        nest_stack_id = vnf_instance.instantiated_vnf_info.instance_id
        upf_cp_name = additional_param.get('upf_cp_name')
        upf_config_path = additional_param.get('upf_config_path')
        if not upf_config_path:
            upf_config_path = '~/free5gc/src/upf/build/config/upfcfg.yaml'
        start_process_script_path = \
            additional_param.get('start_process_script_path')
        if not start_process_script_path:
            start_process_script_path = '~/free5gc/run.sh'
            start_process_script_dir = \
                start_process_script_path.replace('/run.sh', '')
        smf_config_path = additional_param.get('smf_config_path')
        if not smf_config_path:
            smf_config_path = '~/free5gc/config/smfcfg.conf'
        vim_connection_info = \
            self._get_vim_connection_info(context, vnf_instance)
        heatclient = hc.HeatClient(vim_connection_info.access_info)
        scale_out_id_list = kwargs.get('scale_out_id_list')
        group_resources_list = self._get_group_resources_list(
            heatclient, nest_stack_id, scale_vnf_request.aspect_id,
            vnf_instance.instantiated_vnf_info.additional_params)
        for group_resource in group_resources_list:
            stack_id = group_resource.physical_resource_id
            if stack_id not in scale_out_id_list:
                resource_name = ssh_cp_name
                resource_info = heatclient.resources.get(
                    stack_id=stack_id,
                    resource_name=resource_name)
                # get ssh_ip
                if resource_info.attributes.get('floating_ip_address'):
                    vnf1_ssh_ip = resource_info.attributes.get(
                        'floating_ip_address')
                else:
                    vnf1_ssh_ip = heatclient.resources.get(
                        stack_id=stack_id,
                        resource_name=resource_name).attributes.get(
                        'fixed_ips')[0].get('ip_address')
            else:
                resource_name = ssh_cp_name
                resource_info = heatclient.resources.get(
                    stack_id=stack_id,
                    resource_name=resource_name)
                # get ssh_ip
                if resource_info.attributes.get('floating_ip_address'):
                    vnf2_ssh_ip = resource_info.attributes.get(
                        'floating_ip_address')
                else:
                    vnf2_ssh_ip = heatclient.resources.get(
                        stack_id=stack_id,
                        resource_name=resource_name).attributes.get(
                        'fixed_ips')[0].get('ip_address')
                upf_pfcp_ip = heatclient.resources.get(
                    stack_id=stack_id,
                    resource_name=upf_cp_name).attributes.get(
                    'fixed_ips')[0].get('ip_address')

        # stop VNF1's smf process
        commander = cmd_executer.RemoteCommandExecutor(
            user=username, password=password,
            host=vnf1_ssh_ip,
            timeout=30)
        ssh_command = "ps -ef | grep smf | awk '{print $2}'"
        process_pid = self._execute_command(
            commander, ssh_command,
            FREE5GC_CMD_TIMEOUT, 'common', 0)[0].replace('\n', '')
        ssh_command = "kill -9 {}".format(process_pid)
        self._execute_command(
            commander, ssh_command,
            FREE5GC_CMD_TIMEOUT, 'common', 0)
        time.sleep(30)
        commander.close_session()

        # modify upf2's config file
        commander = cmd_executer.RemoteCommandExecutor(
            user=username, password=password,
            host=vnf2_ssh_ip,
            timeout=30)
        ssh_command = ('sed -i "s/addr: 172.168.151.92/addr: {upf_pfcp_ip}/g"'
                       ' {upf_config_path}'
                       .format(upf_pfcp_ip=upf_pfcp_ip,
                               upf_config_path=upf_config_path))
        self._execute_command(
            commander, ssh_command,
            FREE5GC_CMD_TIMEOUT, 'common', 0)
        # start upf2
        ssh_command = 'cd {};cd src/upf/build;sudo -E ./bin/free5gc-upfd &'.\
            format(start_process_script_dir)
        self._execute_command(
            commander, ssh_command,
            FREE5GC_CMD_TIMEOUT, 'common', 0)
        commander.close_session()

        # add new upf's config in smf config
        commander = cmd_executer.RemoteCommandExecutor(
            user=username, password=password,
            host=vnf1_ssh_ip,
            timeout=30)
        ssh_command = r"sed -i '/^        node_id:*/a\      UPF2:\\n" \
                      "        type: UPF\\n" \
                      "        node_id: {}' {}".\
            format(upf_pfcp_ip, smf_config_path)
        self._execute_command(
            commander, ssh_command,
            FREE5GC_CMD_TIMEOUT, 'common', 0)
        # start smf process
        ssh_command = "cd {};" \
                      "touch smf.txt;".format(start_process_script_dir)
        self._execute_command(
            commander, ssh_command,
            FREE5GC_CMD_TIMEOUT, 'common', 0)
        ssh_command = "cd {};echo './bin/smf &' | tee run_smf.sh;" \
                      "chmod 777 run_smf.sh".format(start_process_script_dir)
        self._execute_command(
            commander, ssh_command,
            FREE5GC_CMD_TIMEOUT, 'common', 0)
        ssh_command = "cd {};nohup ./run_smf.sh > smf.txt 2>&1 &".format(
            start_process_script_dir)
        self._execute_command(
            commander, ssh_command,
            FREE5GC_CMD_TIMEOUT, 'common', 0)
        commander.close_session()

    def _get_vnfc_resource_id(self, vnfc_resource_info, vnfc_instance_id):
        for vnfc_resource in vnfc_resource_info:
            if vnfc_resource.id == vnfc_instance_id:
                return vnfc_resource
        else:
            return None

    def _get_target_physical_resource_ids(self, vnf_instance,
                                          heal_vnf_request):
        target_physical_resource_ids = []
        for vnfc_instance_id in heal_vnf_request.vnfc_instance_id:
            instantiated_vnf_info = vnf_instance.instantiated_vnf_info
            vnfc_resource_info = instantiated_vnf_info.vnfc_resource_info
            vnfc_resource = self._get_vnfc_resource_id(
                vnfc_resource_info, vnfc_instance_id)
            if vnfc_resource:
                target_physical_resource_ids.append(
                    vnfc_resource.compute_resource.resource_id)

        return target_physical_resource_ids

    @log.log
    def heal_start(self, context, vnf_instance,
                   heal_vnf_request, grant,
                   grant_request, **kwargs):
        if heal_vnf_request.vnfc_instance_id:
            additional_param = vnf_instance.instantiated_vnf_info. \
                additional_params.get('free5gc', {})
            aspect_id = additional_param.get('aspect_id')
            ssh_cp_name = additional_param.get('ssh_cp_name')
            upf_cp_name = additional_param.get('upf_cp_name')
            vim_connection_info = \
                self._get_vim_connection_info(context, vnf_instance)
            heatclient = hc.HeatClient(vim_connection_info.access_info)
            nest_stack_id = vnf_instance.instantiated_vnf_info.instance_id
            group_resources_list = self._get_group_resources_list(
                heatclient, nest_stack_id, aspect_id,
                vnf_instance.instantiated_vnf_info.additional_params)
            if len(group_resources_list) > 1:
                target_physical_resource_ids = \
                    self._get_target_physical_resource_ids(
                        vnf_instance, heal_vnf_request)
                vnf_ssh_ip, _ = self._get_vnf_ssh_ip(
                    group_resources_list, heatclient, ssh_cp_name,
                    target_physical_resource_ids, upf_cp_name)
                vnf1_ssh_ip = vnf_ssh_ip['not_heal']
                # connect to vnf1
                username = additional_param.get('username')
                password = additional_param.get('password')
                commander = cmd_executer.RemoteCommandExecutor(
                    user=username, password=password,
                    host=vnf1_ssh_ip,
                    timeout=30)
                ssh_command = r"ps -ef | grep '\./bin' | grep smf | " \
                              "awk '{print $2}'"
                # stop smf's process
                process = self._execute_command(
                    commander, ssh_command,
                    FREE5GC_CMD_TIMEOUT, 'common', 0)
                if len(process) > 1:
                    process_pid = process[0].replace('\n', '')
                    ssh_command = "kill -9 {}".format(process_pid)
                    self._execute_command(
                        commander, ssh_command,
                        FREE5GC_CMD_TIMEOUT, 'common', 0)
                    time.sleep(30)
                commander.close_session()

    def _get_cp_ip(self, ssh_cp_name, heatclient, stack_id):
        resource_name = ssh_cp_name
        resource_info = heatclient.resources.get(
            stack_id=stack_id,
            resource_name=resource_name)
        # get ssh_ip
        if resource_info.attributes.get('floating_ip_address'):
            cp_ip = resource_info.attributes.get(
                'floating_ip_address')
        else:
            cp_ip = heatclient.resources.get(
                stack_id=stack_id,
                resource_name=resource_name).attributes.get(
                'fixed_ips')[0].get('ip_address')
        return cp_ip

    def _get_vnf_ssh_ip(self, group_resources_list, heatclient,
                        ssh_cp_name, target_physical_resource_ids,
                        upf_cp_name):
        vnf_ssh_ip = {}
        for group_resource in group_resources_list:
            stack_id = group_resource.physical_resource_id
            resource_info_list = heatclient.resources.list(
                stack_id=stack_id)
            for resource in resource_info_list:
                if resource.resource_type == 'OS::Nova::Server' and \
                        resource.physical_resource_id not in \
                        target_physical_resource_ids:
                    vnf_ssh_ip['not_heal'] = self._get_cp_ip(
                        ssh_cp_name, heatclient, stack_id)
                if resource.resource_type == 'OS::Nova::Server' and \
                        resource.physical_resource_id in \
                        target_physical_resource_ids:
                    vnf_ssh_ip['heal'] = self._get_cp_ip(
                        ssh_cp_name, heatclient, stack_id)
                    vnf_ssh_ip['upf_pfcp_ip'] = heatclient.resources.get(
                        stack_id=stack_id,
                        resource_name=upf_cp_name).attributes.get(
                        'fixed_ips')[0].get('ip_address')
                    heal_stack_id = stack_id

        return vnf_ssh_ip, heal_stack_id

    @log.log
    def heal_end(self, context, vnf_instance,
                 heal_vnf_request, grant,
                 grant_request, **kwargs):
        nest_stack_id = vnf_instance.instantiated_vnf_info.instance_id
        additional_param = vnf_instance.instantiated_vnf_info. \
            additional_params.get('free5gc', {})
        ssh_cp_name = additional_param.get('ssh_cp_name')
        username = additional_param.get('username')
        password = additional_param.get('password')
        upf_cp_name = additional_param.get('upf_cp_name')
        upf_config_path = additional_param.get('upf_config_path')
        if not upf_config_path:
            upf_config_path = '~/free5gc/src/upf/build/config/upfcfg.yaml'
        start_process_script_path = \
            additional_param.get('start_process_script_path')
        if not start_process_script_path:
            start_process_script_path = '~/free5gc/run.sh'
            start_process_script_dir = \
                start_process_script_path.replace('/run.sh', '')
        vim_connection_info = \
            self._get_vim_connection_info(context, vnf_instance)
        heatclient = hc.HeatClient(vim_connection_info.access_info)
        aspect_id = additional_param.get('aspect_id')
        group_resources_list = self._get_group_resources_list(
            heatclient, nest_stack_id, aspect_id,
            vnf_instance.instantiated_vnf_info.additional_params)

        if not heal_vnf_request.vnfc_instance_id or len(
                group_resources_list) == 1:
            self.instantiate_end(
                context, vnf_instance,
                vnf_instance.instantiated_vnf_info,
                grant, grant_request, **kwargs)
        else:
            target_physical_resource_ids = \
                self._get_target_physical_resource_ids(
                    vnf_instance, heal_vnf_request)
            vnf_ssh_ip, _ = self._get_vnf_ssh_ip(
                group_resources_list, heatclient, ssh_cp_name,
                target_physical_resource_ids, upf_cp_name)
            vnf_not_heal_ssh_ip = vnf_ssh_ip['not_heal']
            commander = cmd_executer.RemoteCommandExecutor(
                user=username, password=password,
                host=vnf_not_heal_ssh_ip,
                timeout=30)
            ssh_command = r"ps -ef | grep '\./bin' | grep amf | " \
                          "awk '{print $2}'"
            process = self._execute_command(
                commander, ssh_command,
                FREE5GC_CMD_TIMEOUT, 'common', 0)
            if len(process) > 1:
                healed_VM_smf_flag = False

            if not healed_VM_smf_flag:
                vnf2_ssh_ip = vnf_ssh_ip['heal']
                vnf1_ssh_ip = vnf_ssh_ip['not_heal']
                upf_pfcp_ip = vnf_ssh_ip['upf_pfcp_ip']

                # modify upf2's config file and start upf2
                commander = cmd_executer.RemoteCommandExecutor(
                    user=username, password=password,
                    host=vnf2_ssh_ip,
                    timeout=30)
                ssh_command = \
                    'sed -i "s/addr: 172.168.151.92/addr: {upf_pfcp_ip}"' \
                    ' {upf_config_path}'.format(
                        upf_pfcp_ip=upf_pfcp_ip,
                        upf_config_path=upf_config_path)
                self._execute_command(
                    commander, ssh_command,
                    FREE5GC_CMD_TIMEOUT, 'common', 0)
                # start upf2
                ssh_command = 'cd {};cd src/upf/build;' \
                              'sudo -E ./bin/free5gc-upfd &'. \
                    format(start_process_script_dir)
                self._execute_command(
                    commander, ssh_command,
                    FREE5GC_CMD_TIMEOUT, 'common', 0)
                commander.close_session()

                # start vnf1's smf's process
                commander = cmd_executer.RemoteCommandExecutor(
                    user=username, password=password,
                    host=vnf1_ssh_ip,
                    timeout=30)
                ssh_command = "cd {};" \
                              "touch smf.txt;".format(start_process_script_dir)
                self._execute_command(
                    commander, ssh_command,
                    FREE5GC_CMD_TIMEOUT, 'common', 0)
                ssh_command = ("cd {};echo './bin/smf &' | tee run_smf.sh;"
                               "chmod 777 run_smf.sh"
                               .format(start_process_script_dir))
                self._execute_command(
                    commander, ssh_command,
                    FREE5GC_CMD_TIMEOUT, 'common', 0)
                ssh_command = ("cd {};nohup ./run_smf.sh > smf.txt 2>&1 "
                               "&".format(start_process_script_dir))
                self._execute_command(
                    commander, ssh_command,
                    FREE5GC_CMD_TIMEOUT, 'common', 0)
                commander.close_session()
            else:
                vnf_package_path = vnflcm_utils._get_vnf_package_path(
                    context, vnf_instance.vnfd_id)
                self._modify_config_path_and_start_process(
                    group_resources_list, heatclient, vnf_package_path,
                    additional_param, 'HEAL',
                    target_physical_resource_ids=target_physical_resource_ids)

    @log.log
    def change_external_connectivity_start(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass

    @log.log
    def change_external_connectivity_end(
            self, context, vnf_instance,
            change_ext_conn_request, grant,
            grant_request, **kwargs):
        pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heatclient.v1 import resources
from tacker.tests.unit import base
from tacker.tests import uuidsentinel
from tacker.vnfm.infra_drivers.openstack.heat_client import HeatClient
from tacker.vnfm.infra_drivers.openstack.update_template import HOTUpdater


class TestUpdateTemplate(base.TestCase):

    def setUp(self):
        super(TestUpdateTemplate, self).setUp()
        self.maxDiff = None
        self._mock_heatclient()
        self.stack_id = uuidsentinel.stack_id
        self.hot_updater = HOTUpdater(self.heatclient)

    def _mock_heatclient(self):
        self.heatclient = mock.Mock(spec=HeatClient)
        self.heatclient.stacks = mock.Mock()
        self.heatclient.stacks.template.side_effect = [
            self._get_template(),
            self._get_intermediate_template1(),
            self._get_nested_template1(),
            self._get_intermediate_template2(),
            self._get_nested_template2()
        ]
        self.heatclient.resource_get_list.return_value = \
            self._get_stack_resources()

    def test_get_templates_from_stack(self):
        self.hot_updater.get_templates_from_stack(self.stack_id)

        self.assertDictEqual(self._get_template(), self.hot_updater.template)
        self.assertDictEqual(
            {'VDU1.yaml': self._get_nested_template1(),
             'VDU2.yaml': self._get_nested_template2()},
            self.hot_updater.nested_templates)

    def test_update_resource_property(self):
        self.hot_updater.get_templates_from_stack(self.stack_id)

        # Test pattern 1: Change network and CP1(DHCP) address to fixed.
        self.hot_updater.update_resource_property(
            'CP1', ['OS::Neutron::Port'],
            network=uuidsentinel.network_id_cp1_changed,
            fixed_ips=[{'ip_address': '20.20.0.100'}])
        expected = {
            'network': uuidsentinel.network_id_cp1_changed,
            'fixed_ips': [{'ip_address': '20.20.0.100'}]
        }
        self.assertEqual(
            expected,
            self.hot_updater.template['resources']['CP1']['properties'])

        # Test pattern 2: Change network and fixed address.
        self.hot_updater.update_resource_property(
            'CP2', ['OS::Neutron::Port'],
            network=uuidsentinel.network_id_cp2_changed,
            fixed_ips=[{
                'ip_address': '20.20.0.200',
                'subnet': uuidsentinel.subnet_id_cp2_changed
            }])
        expected = {
            'network': uuidsentinel.network_id_cp2_changed,
            'fixed_ips': [{
                'ip_address': '20.20.0.200',
                'subnet': uuidsentinel.subnet_id_cp2_changed,
            }]
        }
        self.assertEqual(
            expected,
            self.hot_updater.template['resources']['CP2']['properties'])

        # Test Pattern 3: Delete fixed_ips property.
        self.hot_updater.update_resource_property(
            'CP3', ['OS::Neutron::Port'],
            network=uuidsentinel.network_id_cp3_changed,
            fixed_ips=None)
        self.assertIsNone(self.hot_updater.nested_templates[
            'VDU1.yaml']['resources']['CP3']['properties'].get('fixed_ips'))

    def test_update_resource_property_not_found(self):
        self.hot_updater.get_templates_from_stack(self.stack_id)

        # Test Pattern 4: Resource does not exist.
        self.hot_updater.update_resource_property(
            'CPX', ['OS::Neutron::Port'],
            network=uuidsentinel.network_id_cpx,
            fixed_ips=[{'ip_address': '20.20.0.100'}])
        self.assertEqual(self._get_template(), self.hot_updater.template)

        # Test Pattern 5: Resource type does not exist.
        self.hot_updater.update_resource_property(
            'CP1', ['OS::Sahara::Cluster'],
            network=uuidsentinel.network_id_cp1_changed,
            fixed_ips=[{'ip_address': '20.20.0.100'}])
        self.assertEqual(self._get_template(), self.hot_updater.template)

        # Test Pattern 6: Resource doess not have properties.
        self.hot_updater.update_resource_property(
            'CP5', ['OS::Neutron::Port'],
            network=uuidsentinel.network_id_cp5,
            fixed_ips=[{'ip_address': '20.20.0.100'}])
        self.assertEqual(self._get_template(), self.hot_updater.template)

    def _get_template(self):
        return {
            'heat_template_version': '2013-05-23',
            'description': 'Simple deployment flavour for Sample VNF',
            'parameters': {},
            'resources': {
                'CP1': {
                    'type': 'OS::Neutron::Port',
                    'properties': {
                        'network': uuidsentinel.network_id_cp1,
                    },
                },
                'CP2': {
                    'type': 'OS::Neutron::Port',
                    'properties': {
                        'network': uuidsentinel.network_id_cp2,
                        'fixed_ips': [{
                            'ip_address': '10.10.0.200',
                            'subnet': uuidsentinel.subnet_id_cp2,
                        }],
                    },
                },
                'CP5': {
                    'type': 'OS::Neutron::Port',
                },
            },
        }

    def _get_intermediate_template1(self):
        return {
            'heat_template_version': '2013-05-23',
            'description': 'Simple deployment flavour for Sample VNF',
            'parameters': {},
            'resources': {
                'xxxxxxxx': {
                    'type': 'VDU1.yaml',
                },
            },
        }

    def _get_intermediate_template2(self):
        return {
            'heat_template_version': '2013-05-23',
            'description': 'Simple deployment flavour for Sample VNF',
            'parameters': {},
            'resources': {
                'yyyyyyyy': {
                    'type': 'VDU2.yaml',
                },
            },
        }

    def _get_nested_template1(self):
        return {
            'heat_template_version': '2013-05-23',
            'description': 'Simple deployment flavour for Sample VNF',
            'parameters': {},
            'resources': {
                'CP3': {
                    'type': 'OS::Neutron::Port',
                    'properties': {
                        'network': uuidsentinel.network_id_cp3,
                        'fixed_ips': [{
                            'subnet': uuidsentinel.subnet_id_cp3,
                        }],
                    },
                },
            },
        }

    def _get_nested_template2(self):
        return {
            'heat_template_version': '2013-05-23',
            'description': 'Simple deployment flavour for Sample VNF',
            'parameters': {},
            'resources': {
                'CP4': {
                    'type': 'OS::Neutron::Port',
                    'properties': {
                        'network': uuidsentinel.network_id_cp4,
                        'fixed_ips': [{
                            'subnet': uuidsentinel.subnet_id_cp4,
                        }],
                    },
                },
            },
        }

    def _get_stack_resources(self):
        def _create_resource(resource_name, resource_type):
            return resources.Resource(None, {
                'resource_name': resource_name,
                'resource_type': resource_type,
                'resource_status': 'CREATE_COMPLETE',
                'physical_resource_id': uuidsentinel.uuid,
            })

        data = [
            ('VDU1_scale_group', 'OS::Heat::AutoScalingGroup'),
            ('xxxxxxxx', 'VDU1.yaml'),
            ('VDU2_scale_group', 'OS::Heat::AutoScalingGroup'),
            ('yyyyyyyy', 'VDU2.yaml'),
            ('CP1', 'OS::Neutron::Port'),
            ('CP2', 'OS::Neutron::Port'),
            ('CP3', 'OS::Neutron::Port'),
            ('CP4', 'OS::Neutron::Port')
        ]
        return [_create_resource(row[0], row[1]) for row in data]



# Copyright 2013, 2014 Intel Corporation.
# All Rights Reserved.
#
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from tacker.vnfm.mgmt_drivers import abstract_driver


LOG = logging.getLogger(__name__)


class VnfMgmtNoop(abstract_driver.VnfMGMTAbstractDriver):
    def get_type(self):
        return 'noop'

    def get_name(self):
        return 'noop'

    def get_description(self):
        return 'Tacker VNFMgmt Noop Driver'

    def mgmt_ip_address(self, plugin, context, vnf):
        LOG.debug('mgmt_ip_address %s', vnf)
        return 'noop-mgmt-url'

    def mgmt_call(self, plugin, context, vnf, kwargs):
        LOG.debug('mgmt_call %(vnf)s %(kwargs)s',
                  {'vnf': vnf, 'kwargs': kwargs})



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ast

from tacker.mistral import workflow_generator


OUTPUT = {
    'create_vnf': ['vnf_id', 'vim_id', 'mgmt_ip_address', 'status'],
    'create_vnffg': ['vnffg_id'],
}


class WorkflowGenerator(workflow_generator.WorkflowGeneratorBase):

    def _add_create_vnf_tasks(self, ns):
        vnfds = ns['vnfd_details']
        task_dict = dict()
        for vnfd_name, vnfd_info in (vnfds).items():
            nodes = vnfd_info['instances']
            for node in nodes:
                task = self.wf_name + '_' + node
                task_dict[task] = {
                    'action': 'tacker.create_vnf body=<% $.ns.{0} '
                              '%>'.format(node),
                    'input': {'body': '<% $.ns.{0} %>'.format(node)},
                    'publish': {
                        'vnf_id_' + node: '<% task({0}).result.vnf.id '
                                          '%>'.format(task),
                        'vim_id_' + node: '<% task({0}).result.vnf.vim_id'
                                          ' %>'.format(task),
                        'mgmt_ip_address_' + node: '<% task({0}).result.vnf.'
                                                   'mgmt_ip_address '
                                                   '%>'.format(task),
                        'status_' + node: '<% task({0}).result.vnf.status'
                                          ' %>'.format(task),
                              },
                    'on-success': ['wait_vnf_active_%s' % node]
                }
        return task_dict

    def _add_wait_vnf_tasks(self, ns):
        vnfds = ns['vnfd_details']
        task_dict = dict()
        for vnfd_name, vnfd_info in (vnfds).items():
            nodes = vnfd_info['instances']
            for node in nodes:
                task = 'wait_vnf_active_%s' % node
                task_dict[task] = {
                    'action': 'tacker.show_vnf vnf=<% $.vnf_id_{0} '
                              '%>'.format(node),
                    'retry': {
                        'count': 10,
                        'delay': 10,
                        'break-on': '<% $.status_{0} = "ERROR"'
                                    ' %>'.format(node),
                        'continue-on': '<% $.status_{0} = "PENDING_CREATE" '
                                       '%>'.format(node),
                              },
                    'publish': {
                        'mgmt_ip_address_' + node: ' <% task({0}).result.'
                                                   'vnf.mgmt_ip_address '
                                                   '%>'.format(task),
                        'status_' + node: '<% task({0}).result.vnf.status'
                                          ' %>'.format(task),
                              },
                    'on-success': [
                        {'delete_vnf_' + node: '<% $.status_{0}='
                                               '"ERROR" %>'.format(node)}
                              ]
                }
                vnffgd_templates = ns.get('vnffgd_templates')
                if vnffgd_templates:
                    for vnffg_name in vnffgd_templates:
                        vnffg_group = vnffgd_templates[vnffg_name][
                            'topology_template']['groups'][vnffg_name]
                        constituent_vnfs = vnffg_group[
                            'properties']['constituent_vnfs']

                        if vnfd_name in constituent_vnfs:
                            task_dict[task]['on-success'].append(
                                'create_vnffg_%s' % vnffg_name)
        return task_dict

    def _add_delete_vnf_tasks(self, ns, vnffg_ids=None):
        vnfds = ns['vnfd_details']
        vnf_attr = {'vnf': {'attributes': {
                            'force': ns.get('force_delete', False)}}}
        task_dict = dict()
        for vnfd_name, vnfd_info in (vnfds).items():
            nodes = vnfd_info['instances']
            for node in nodes:
                task = 'delete_vnf_%s' % node
                task_dict[task] = {
                    'action': 'tacker.delete_vnf vnf=<% $.vnf_id_{0}'
                              '%>'.format(node),
                    'input': {'body': vnf_attr},
                }
                if vnffg_ids and len(vnffg_ids):
                    task_dict[task].update({'join': 'all'})
        return task_dict

    def _add_create_vnffg_task(self, vnffgd_templates):
        task_dict = dict()
        previous_task = None
        for vnffg_name in vnffgd_templates:
            task = 'create_vnffg_%s' % vnffg_name
            vnffg_output = 'vnffg_id_%s' % vnffg_name
            task_dict[task] = {
                'join': 'all',
                'action': 'tacker.create_vnffg body=<% $.ns.{0} '
                          '%>'.format(vnffg_name),
                'input': {'body': '<% $.ns.{0} %>'.format(vnffg_name)},
                'publish': {
                    vnffg_output: '<% task({0}).result.'
                                  'vnffg.id %>'.format(task)}
            }
            if previous_task:
                task_dict[previous_task].update({'on-success': [task]})
            previous_task = task
        return task_dict

    def _add_delete_vnffg_task(self, ns):
        task_dict = dict()
        vnfds = ns['vnfd_details']
        vnffg_ids = ns['vnffg_details']
        delayed_tasks = list()

        for vnfd_name, vnfd_info in (vnfds).items():
            nodes = vnfd_info['instances']
            for node in nodes:
                wait_task = 'delete_vnf_%s' % node
                delayed_tasks.append(wait_task)

        previous_task = None
        for vnffg_name in vnffg_ids:
            task = 'delete_vnffg_%s' % vnffg_name
            if previous_task:
                wait_tasks = delayed_tasks + [previous_task]
            else:
                wait_tasks = delayed_tasks
            previous_task = task
            task_dict[task] = {
                'action': 'tacker.delete_vnffg vnffg=<% $.{0} '
                          '%>'.format(vnffg_name),
                'on-success': wait_tasks
            }
        return task_dict

    def _build_output_dict(self, ns):
        vnfds = ns['vnfd_details']
        task_dict = dict()
        for vnfd_name, vnfd_info in (vnfds).items():
            nodes = vnfd_info['instances']
            for node in nodes:
                for op_name in OUTPUT['create_vnf']:
                    task_dict[op_name + '_' + node] = \
                        '<% $.{0}_{1} %>'.format(op_name, node)
        vnffgd_templates = ns.get('vnffgd_templates')
        if vnffgd_templates:
            for vnffg_name in vnffgd_templates:
                for op_name in OUTPUT['create_vnffg']:
                    vnffg_output = '%s_%s' % (op_name, vnffg_name)
                    task_dict[vnffg_output] = \
                        '<% $.{0}_{1} %>'.format(op_name, vnffg_name)
        return task_dict

    def get_input_dict(self):
        return self.input_dict

    def build_input(self, ns, params):
        vnfds = ns['vnfd_details']
        ns_id = ns['ns'].get('ns_id')
        ns_name = ns['ns'].get('name')
        self.input_dict = {'ns': {}}
        for vnfd_name, vnfd_info in (vnfds).items():
            nodes = vnfd_info['instances']
            for node in nodes:
                vnf_name = '%s_VNF_%s' % (ns_name, vnfd_info['id'])
                self.input_dict['ns'][node] = dict()
                self.input_dict['ns'][node]['vnf'] = {
                    'attributes': {},
                    'vim_id': ns['ns'].get('vim_id', ''),
                    'vnfd_id': vnfd_info['id'],
                    'name': vnf_name
                }
                if params.get(vnfd_name):
                    self.input_dict['ns'][node]['vnf']['attributes'] = {
                        'param_values': params.get(vnfd_name)
                    }
        if ns.get('vnffgd_templates'):
            vnffg_input = self.build_vnffg_input(ns, params, ns_id)
            self.input_dict['ns'].update(vnffg_input)

    def build_vnffg_input(self, ns, params, ns_id):
        vnffgd_templates = ns.get('vnffgd_templates')
        vnffg_input = dict()
        for vnffg_name in vnffgd_templates:
            vnffg_group = vnffgd_templates[vnffg_name][
                'topology_template']['groups'][vnffg_name]
            constituent_vnfs = vnffg_group[
                'properties']['constituent_vnfs']
            vnf_mapping = self.get_vnf_mapping(ns, constituent_vnfs)
            vnffgd_body = dict()
            vnffgd_body['vnffg'] = {
                'name': '%s_%s_%s' % (ns['ns'].get('name'), vnffg_name, ns_id),
                'vnffgd_template': vnffgd_templates[vnffg_name],
                'vnf_mapping': vnf_mapping,
                'attributes': {
                    'param_values': params.get('nsd')},
                'ns_id': ns_id
            }
            vnffg_input[vnffg_name] = vnffgd_body
        return vnffg_input

    def get_vnf_mapping(self, ns, constituent_vnfs):
        vnfds = ns['vnfd_details']
        vnf_mapping = dict()
        for vnfd_name, vnfd_info in (vnfds).items():
            if vnfd_name in constituent_vnfs:
                vnf_name = '%s_VNF_%s' % (ns['ns'].get('name'),
                                          vnfd_info['id'])
                vnf_mapping[vnfd_name] = vnf_name
        return vnf_mapping

    def create_ns(self, **kwargs):
        ns = kwargs.get('ns')
        params = kwargs.get('params')

        tasks = {}
        for func in [self._add_create_vnf_tasks,
                     self._add_wait_vnf_tasks,
                     self._add_delete_vnf_tasks]:
            tasks.update(func(ns))

        self.build_input(ns, params)
        vnffgd_templates = ns.get('vnffgd_templates')
        if vnffgd_templates:
            create_task = self._add_create_vnffg_task(vnffgd_templates)
            tasks.update(create_task)

        self.definition[self.wf_identifier]['tasks'] = tasks
        self.definition[self.wf_identifier]['output'] = \
            self._build_output_dict(ns)

    def delete_ns(self, ns):
        ns_dict = {'vnfd_details': {}}
        vnf_ids = ast.literal_eval(ns['vnf_ids'])
        self.definition[self.wf_identifier]['input'] = []
        for vnf in vnf_ids:
            vnf_key = 'vnf_id_' + vnf
            self.definition[self.wf_identifier]['input'].append(vnf_key)
            self.input_dict[vnf_key] = vnf_ids[vnf]
            ns_dict['vnfd_details'][vnf] = {'instances': [vnf]}
        self.definition[self.wf_identifier]['tasks'] = dict()

        vnffg_ids = ast.literal_eval(ns.get('vnffg_ids'))
        if len(vnffg_ids):
            for vnffg_name in vnffg_ids:
                self.definition[self.wf_identifier]['input'].append(vnffg_name)
                self.input_dict[vnffg_name] = vnffg_ids[vnffg_name]
                ns_dict['vnffg_details'] = vnffg_ids
            self.definition[self.wf_identifier]['tasks'].update(
                self._add_delete_vnffg_task(ns_dict))
        ns_dict['force_delete'] = ns.get('force_delete', False)
        self.definition[self.wf_identifier]['tasks'].update(
            self._add_delete_vnf_tasks(ns_dict, vnffg_ids))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy.orm import exc as orm_exc
from unittest import mock

from tacker.extensions import nfvo
from tacker import manager
from tacker.tests.unit import base
from tacker.vnfm import vim_client


class TestVIMClient(base.TestCase):

    def setUp(self):
        super(TestVIMClient, self).setUp()
        self.vim_info = {'id': 'aaaa', 'name': 'VIM0', 'type': 'test_vim',
                         'auth_cred': {'password': '****'},
                         'auth_url': 'http://127.0.0.1/identity/v3',
                         'placement_attr': {'regions': ['TestRegionOne']},
                         'tenant_id': 'test'}
        self.vimclient = vim_client.VimClient()
        self.service_plugins = mock.Mock()
        self.nfvo_plugin = mock.Mock()

    def test_get_vim_without_defined_default_vim(self):
        self.nfvo_plugin.get_default_vim.side_effect = \
            orm_exc.NoResultFound()
        self.service_plugins.get.return_value = self.nfvo_plugin
        with mock.patch.object(manager.TackerManager, 'get_service_plugins',
                               return_value=self.service_plugins):
            self.assertRaises(nfvo.VimDefaultNotDefined,
                              self.vimclient.get_vim, None)

    def test_get_vim_not_found_exception(self):
        vim_id = self.vim_info['id']
        self.nfvo_plugin.get_vim.side_effect = \
            orm_exc.NoResultFound()
        self.service_plugins.get.return_value = self.nfvo_plugin
        with mock.patch.object(manager.TackerManager, 'get_service_plugins',
                               return_value=self.service_plugins):
            self.assertRaises(nfvo.VimNotFoundException,
                              self.vimclient.get_vim, None, vim_id=vim_id)

    def test_get_vim_region_not_found_region_name_invalid(self):
        self.nfvo_plugin.get_vim.return_value = self.vim_info
        self.service_plugins.get.return_value = self.nfvo_plugin
        with mock.patch.object(manager.TackerManager, 'get_service_plugins',
                               return_value=self.service_plugins):
            self.assertRaises(nfvo.VimRegionNotFoundException,
                              self.vimclient.get_vim, None,
                              vim_id=self.vim_info['id'],
                              region_name='Test')

    def test_get_vim(self):
        self.nfvo_plugin.get_vim.return_value = self.vim_info
        self.service_plugins.get.return_value = self.nfvo_plugin
        self.vimclient._build_vim_auth = mock.Mock()
        self.vimclient._build_vim_auth.return_value = {'password': '****'}
        with mock.patch.object(manager.TackerManager, 'get_service_plugins',
                               return_value=self.service_plugins):
            vim_result = self.vimclient.get_vim(None,
                                                vim_id=self.vim_info['id'],
                                                region_name='TestRegionOne')
            vim_expect = {'vim_auth': {'password': '****'}, 'vim_id': 'aaaa',
                          'vim_name': 'VIM0', 'vim_type': 'test_vim',
                          'placement_attr': {'regions': ['TestRegionOne']},
                          'tenant': 'test', 'extra': {}}
            self.assertEqual(vim_expect, vim_result)

    def test_get_vim_with_default_name(self):
        self.vim_info.pop('name')
        self.nfvo_plugin.get_vim.return_value = self.vim_info
        self.service_plugins.get.return_value = self.nfvo_plugin
        self.vimclient._build_vim_auth = mock.Mock()
        self.vimclient._build_vim_auth.return_value = {'password': '****'}
        with mock.patch.object(manager.TackerManager, 'get_service_plugins',
                               return_value=self.service_plugins):
            vim_result = self.vimclient.get_vim(None,
                                                vim_id=self.vim_info['id'],
                                                region_name='TestRegionOne')
            vim_expect = {'vim_auth': {'password': '****'}, 'vim_id': 'aaaa',
                          'vim_name': 'aaaa', 'vim_type': 'test_vim',
                          'placement_attr': {'regions': ['TestRegionOne']},
                          'tenant': 'test', 'extra': {}}
            self.assertEqual(vim_expect, vim_result)

    def test_find_vim_key_with_key_not_found_exception(self):
        vim_id = self.vim_info['id']
        self.assertRaises(nfvo.VimKeyNotFoundException,
                          self.vimclient._find_vim_key, vim_id)

    def test_region_valid_false(self):
        vim_regions = ['TestRegionOne', 'TestRegionTwo']
        region_name = 'TestRegion'
        self.assertFalse(self.vimclient.region_valid(vim_regions,
                                                     region_name))

    def test_region_valid_true(self):
        vim_regions = ['TestRegionOne', 'TestRegionTwo']
        region_name = 'TestRegionOne'
        self.assertTrue(self.vimclient.region_valid(vim_regions, region_name))



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import threading
import time

from oslo_utils import uuidutils

from tacker.sol_refactored.common import coordinate
from tacker.sol_refactored.common import exceptions as sol_ex
from tacker.tests import base


class TestCoordinate(base.BaseTestCase):

    def setUp(self):
        super(TestCoordinate, self).setUp()
        self.sem_1 = threading.Semaphore(value=0)
        self.sem_2 = threading.Semaphore(value=0)
        self.ok = False

    @coordinate.lock_vnf_instance('{inst_id}')
    def _work_thread_1(self, inst_id, sleep_time):
        # notify to parent
        self.sem_1.release()

        # wait to notify from parent
        self.sem_2.acquire()

        if sleep_time:
            time.sleep(sleep_time)

    @coordinate.lock_vnf_instance('{inst_id}')
    def _work_thread_2(self, inst_id):
        pass

    @coordinate.lock_vnf_instance('{inst_id}', delay=True)
    def _work_thread_3(self, inst_id):
        self.ok = True

    def test_lock_vnf_instance(self):
        inst_id = uuidutils.generate_uuid()
        th = threading.Thread(target=self._work_thread_1, args=(inst_id, 0))
        th.start()

        # wait to run _work_thread_1
        self.sem_1.acquire()

        self.assertRaises(sol_ex.OtherOperationInProgress,
            self._work_thread_2, inst_id)

        self.sem_2.release()
        th.join()

    def test_lock_vnf_instance_delay(self):
        inst_id = uuidutils.generate_uuid()
        th = threading.Thread(target=self._work_thread_1, args=(inst_id, 3))
        th.start()

        # wait to run _work_thread_1
        self.sem_1.acquire()

        self.sem_2.release()
        self._work_thread_3(inst_id=inst_id)

        th.join()

        self.assertTrue(self.ok)



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import yaml

from oslo_config import cfg
from oslo_log import log as logging

from tacker.common import rpc
from tacker.mistral.actionrpc import kill_action as killaction
from tacker.mistral import mistral_client
from tacker.nfvo.workflows.vim_monitor import workflow_generator
from tacker.vnfm import keystone


LOG = logging.getLogger(__name__)


def get_mistral_client(auth_dict):
    return mistral_client.MistralClient(
        keystone.Keystone().initialize_client(**auth_dict),
        auth_dict['token']).get_client()


def prepare_and_create_workflow(mistral_client, vim_id, action,
                                kwargs):
    wg = workflow_generator.WorkflowGenerator(vim_id, action)
    wg.task(**kwargs)
    yaml.SafeDumper.ignore_aliases = lambda self, data: True
    definition_yaml = yaml.safe_dump(wg.definition, default_flow_style=False)
    LOG.debug('vim monitor workflow: %s', definition_yaml)
    workflow = mistral_client.workflows.create(definition_yaml)
    return {'id': workflow[0].id, 'input': wg.get_input_dict()}


def execute_workflow(mistral_client, workflow):
    return mistral_client.executions.create(
        wf_identifier=workflow['id'],
        workflow_input=workflow['input'],
        wf_params={})


def delete_executions(mistral_client, vim_id):
    executions = mistral_client.executions.list(
        workflow_name='vim_id_' + vim_id)
    for execution in executions:
        mistral_client.executions.delete(execution.id, force=True)


def delete_workflow(mistral_client, vim_id):
    return mistral_client.workflows.delete('vim_id_' + vim_id)


def monitor_vim(auth_dict, vim_obj):
    mc = get_mistral_client(auth_dict)
    auth_url = vim_obj["auth_url"]
    vim_type = vim_obj['type']
    if vim_type == 'openstack':
        vim_ip = auth_url.split("//")[-1].split(":")[0].split("/")[0]
    elif vim_type == 'kubernetes':
        vim_ip = auth_url.split("//")[-1].split(":")[0]
    workflow_input_dict = {
        'vim_id': vim_obj['id'],
        'count': cfg.CONF.vim_monitor.count,
        'timeout': cfg.CONF.vim_monitor.timeout,
        'interval': cfg.CONF.vim_monitor.interval,
        'targetip': vim_ip}
    workflow = prepare_and_create_workflow(
        mc, vim_obj['id'], 'monitor',
        workflow_input_dict)
    execute_workflow(mc, workflow)


def kill_action(context, vim_obj):
    target = killaction.MistralActionKillRPC.target
    rpc_client = rpc.get_client(target)
    cctxt = rpc_client.prepare(server=vim_obj['id'])
    cctxt.cast(context, 'killAction')


def delete_vim_monitor(context, auth_dict, vim_obj):
    mc = get_mistral_client(auth_dict)
    delete_executions(mc, vim_obj['id'])
    delete_workflow(mc, vim_obj['id'])
    kill_action(context, vim_obj)



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Schema for vnf packages create API.

"""

from tacker.api.validation import parameter_types
from tacker.objects.fields import PackageOperationalStateType

create = {
    'type': 'object',
    'properties': {
        'userDefinedData': parameter_types.keyvalue_pairs
    },
    'additionalProperties': True,
}

upload_from_uri = {
    'type': 'object',
    'properties': {
        'addressInformation': {
            'type': 'string', 'minLength': 0,
            'maxLength': 2048, 'format': 'uri'
        },
        'userName': {
            'type': 'string', 'maxLength': 255,
            'pattern': '^[a-zA-Z0-9-_]*$'
        },
        'password': {
            # Allow to specify any string for strong password.
            'type': 'string', 'maxLength': 255,
        },

    },
    'required': ['addressInformation'],
    'additionalProperties': True,
}

"""
Schema for vnf packages update API.

"""
patch = {
    'type': 'object',
    'properties': {
        'operationalState': {
            'type': 'string',
            'enum': list(PackageOperationalStateType.ALL),
        },
        'userDefinedData': parameter_types.keyvalue_pairs,
    },
    'anyOf': [{'required': ['operationalState']},
              {'required': ['userDefinedData']}],
    'additionalProperties': True
}

query_params_v1 = {
    'type': 'object',
    "properties": {
        'filter': {'type': 'string', 'minLength': 1},
        'exclude_fields': {'type': 'string', 'minLength': 1},
        'fields': {'type': 'string', 'minLength': 1},
        'all_fields': {'format': 'all_fields'},
        'exclude_default': {'format': 'exclude_default'},
    },
    'additionalProperties': True,
}
